import collections
import logging

import pysam
from transgene.common import reject_decision


def reject_snv(snv, rna_bam=None, reject_threshold=None, rna_min_alt_freq=None, dna_bam=None,
               oxog_min_alt_freq=None):
    """
    Decide whether the mutation should be rejected based on the expression filter.

    :param dict snv: A single vcf SNV record split by the tab character and keyed by the fields
    :param str rna_bam: See reject_mutations:`rna_bam`
    :param int reject_threshold: See reject_mutations:`reject_threshold`
    :param float rna_min_alt_freq: See reject_mutations:`rna_min_alt_freq`
    :param str dna_bam: See reject_mutations:`dna_bam`
    :param float oxog_min_alt_freq: See reject_mutation:`oxog_min_alt_freq`

    :return: A named tuple of the True/False if the snv should be rejected or not, and a reason
    for rejection
    :rtype: reject_decision
    """
    bams = {'rna': rna_bam,
            'dna': dna_bam}
    output_counts, reads = get_snv_alignment_info(rna_bam, dna_bam, snv)
    reject = None
    possible_oxog_artefact = False
    low_rna_coverage = None
    rna_alt_freq = None
    coverage = '/'.join([','.join([str(reads[x]['rna']),
                                   str(reads[x]['dna'])]) for x in ('covering', 'spanning')])
    # Very importantly, we first look at dna and then rna
    for bam in 'dna', 'rna':
        # index
        if bams[bam] is None:
            continue
        if reads['spanning'][bam] == 0:
            logging.warning('Mutation at position %s:%s has no coverage in the %s-seq bam. '
                            'Rejecting.', snv['CHROM'], snv['POS'] + 1, bam.upper())
            reject = reject_decision(reject=True, reason='NoReadCoverage', coverage=coverage,
                                     vaf=0.0)
            break
        elif reads['covering'][bam] == 0 and bam == 'rna':
            # This concept of covering vs spanning reads is only in RNA-Seq.
            # I.e. you can have spanning reads but none of them covered the mutation (split).
            logging.warning('Mutation at position %s:%s appears to be in a region that has no '
                            'coverage in the RNA-seq bam, but has %s spanning reads. Possible '
                            'splicing event. Rejecting.', snv['CHROM'], snv['POS'] + 1,
                            reads['spanning'][bam])
            reject = reject_decision(reject=True, reason='SpliceDetected', coverage=coverage,
                                     vaf=0.0)
            break
        elif output_counts[bam][snv['ALT']]['counts'] == 0:
            # This branch will only be hit in RNA-seq cases since 0 Alt will probably never be
            # called by a mutations caller.  If we don't have too many reads over the location, it
            # could either be an undersampled location, or a splice.
            assert bam == 'rna'  # Sanity
            if (reads['covering'][bam] < reject_threshold and
                    reads['spanning'][bam] < reject_threshold):
                if possible_oxog_artefact:
                    # Flagged for RNA rescue but could not be rescued.
                    logging.debug('Mutation at position %s:%s has no evidence of existence in the '
                                  'RNA-seq bam. Possible OxoG variant. rejecting',
                                  snv['CHROM'], snv['POS'] + 1)
                    reject = reject_decision(reject=True, reason='OxoGArtefactLowAlt',
                                             coverage=coverage, vaf=0.0)
                    break
                else:
                    logging.debug('Mutation at position %s:%s has no evidence of existence in the '
                                  'RNA-seq bam. However the coverage at the region (%s/%s '
                                  'reads::Covering/spanning) is below the threshold for rejecting '
                                  '(%s). Accepting.', snv['CHROM'], snv['POS'] + 1,
                                  reads['covering']['rna'], reads['spanning']['rna'],
                                  reject_threshold)
                    low_rna_coverage = True
            else:
                logging.warning('Mutation at position %s:%s has no evidence of existence in the '
                                '%s-seq bam. Coverage = %s/%s reads (Covering/spanning). '
                                'Rejecting.', snv['CHROM'], snv['POS'] + 1, bam.upper(),
                                reads['covering'][bam], reads['spanning'][bam])
                reject = reject_decision(reject=True, reason='NoAltDetected', coverage=coverage,
                                         vaf=0.0)
                break
        else:
            alt_freq = (1.0 * output_counts[bam][snv['ALT']]['counts'] /
                        sum([output_counts[bam][x]['counts'] for x in output_counts[bam]]))
            if bam == 'dna':
                # This can only mean we want OxoG filtering
                if (snv['REF'], snv['ALT']) in (('C', 'A'), ('G', 'T')):
                    if alt_freq <= oxog_min_alt_freq:
                        logging.warning('Mutation at position %s:%s has less than %s ALT allele '
                                        'frequency (%s) in the %s-seq bam. Possible oxoG artefact. '
                                        'Flagging for RNA rescue.', snv['CHROM'], snv['POS'] + 1,
                                        oxog_min_alt_freq, round(alt_freq, 2), bam.upper())
                        possible_oxog_artefact = True
                    elif 0 in [output_counts[bam][snv['ALT']][x] for x in 'fr']:
                        # If either number of 'f'orward or 'r'everse reads is a 0 it means all reads
                        # came from only one strand and hence is possibly an oxoG artefact.
                        if output_counts[bam][snv['ALT']]['f']:
                            dominant_strand = 'forward'
                        else:
                            dominant_strand = 'reverse'
                        logging.warning('Mutation at position %s:%s is called from reads '
                                        'originating only from %s reads in the %s-seq bam. '
                                        'Possible oxoG artefact. Rejecting.', snv['CHROM'],
                                        snv['POS'] + 1, dominant_strand, bam.upper())
                        if dominant_strand == 'forward':
                            reject = reject_decision(reject=True, reason='OxoGArtefactAllFwd',
                                                     coverage=coverage, vaf=0.0)
                        else:
                            reject = reject_decision(reject=True, reason='OxoGArtefactAllRev',
                                                     coverage=coverage, vaf=0.0)
                        break
            else:  # rna
                if alt_freq < rna_min_alt_freq:
                    logging.warning('Mutation at position %s:%s has less than %s ALT allele '
                                    'frequency (%s) in the %s-seq bam. Rejecting.',
                                    snv['CHROM'], snv['POS'] + 1, rna_min_alt_freq,
                                    round(alt_freq, 4), bam.upper())
                    reject = reject_decision(reject=True, reason='LowAltFrequency',
                                             coverage=coverage, vaf=0.0)
                else:
                    rna_alt_freq = alt_freq
    if reject is None:
        # This means the call has not been rejected for any reason. It may have still have been
        # flagged so we need to handle that.
        if possible_oxog_artefact:
            if rna_bam:
                assert rna_alt_freq is not None
                logging.debug('Mutation at position %s:%s with has evidence in the RNA-Seq '
                              '(ALT frequency = %s) despite being flagged as a possible OxoG '
                              'artefect. Accepting', snv['CHROM'], snv['POS'] + 1, rna_alt_freq)
                reject = reject_decision(reject=False, reason='PossibleOxoGArtefactLowAltRNARescue',
                                         coverage=coverage, vaf=rna_alt_freq)
            else:
                logging.warning('Mutation at position %s:%s was flagged as possible oxoG artefact '
                                'and could not be rescued without matching RNA-Seq. Rejecting',
                                snv['CHROM'], snv['POS'] + 1)
                reject = reject_decision(reject=True, reason='OxoGArtefactLowAlt',
                                         coverage=coverage, vaf=0.0)
        else:
            reason = 'LowRNACoverage' if low_rna_coverage else '.'
            logging.debug('Accepted mutation at position %s:%s with %s read coverage and %s VAF.',
                          snv['CHROM'], snv['POS'] + 1, coverage,
                          round(rna_alt_freq, 2) if rna_alt_freq is not None else 'NA')
            reject = reject_decision(reject=False, reason=reason, coverage=coverage,
                                     vaf=0.0 if rna_alt_freq is None else round(rna_alt_freq, 2))
    return reject


def get_snv_alignment_info(rna_bam, dna_bam, vcf_record):
    """
    Get information about the spanning and covering reads at a given genomic locus for an snv.

    :param str rna_bam: Path to the RNA-Seq bam file
    :param str dna_bam: Path to the DNA-Seq bam file
    :param dict vcf_record: A single vcf record split by the tab character and keyed by the fields
    :return: information about the alignment at the position in the vcf record
             Output_counts- The counts of all nucleotides at the position in 'rna' and 'dna', and
             the counts of reads mapping to the 'f'orward and 'r'everse strands for each nucleotide.
             reads - The counts of reads mapping at ('covering') and across ('spanning') the
             position in 'rna' and 'dna'
    :rtype: tuple(collections.Counter|collections.defaultdict(Collections.Counter), dict(dict))
    """
    bams = {'rna': rna_bam, 'dna': dna_bam}
    output_counts = {'rna': collections.Counter(), 'dna': collections.Counter()}
    reads = {'spanning': {'rna': '.', 'dna': '.'},
             'covering': {'rna': '.', 'dna': '.'}}
    for bam in bams:
        if not bams[bam]:
            continue
        samfile = pysam.Samfile(bams[bam], 'rb')
        output_counts_ = collections.defaultdict(collections.Counter)
        covering_reads = 0
        spanning_reads = 0
        pileups = samfile.pileup(vcf_record['CHROM'], vcf_record['POS'], vcf_record['POS'] + 1,
                                 truncate=True)
        for pileup in pileups:
            for read in pileup.pileups:
                if not (read.is_refskip or read.is_del):
                    base = read.alignment.seq[read.query_position]
                    if ord(read.alignment.qual[read.query_position]) >= 63:  # (33PHRED + 30 QUAL)
                        output_counts_[base]['counts'] += 1
                        if read.alignment.is_read2:
                            output_counts_[base]['r'] += 1
                        else:
                            output_counts_[base]['f'] += 1
                    covering_reads += 1
                spanning_reads += 1
        samfile.close()
        reads['spanning'][bam] = spanning_reads
        reads['covering'][bam] = covering_reads
        output_counts[bam] = output_counts_
    return output_counts, reads



import logging

import pysam
from transgene.common import reject_decision, trans


def reject_indel(indel, rna_bam=None, reject_threshold=None, rna_min_alt_freq=None):
    """
    Decide whether the mutation should be rejected based on the expression filter.

    :param dict indel: A single vcf INDEL record split by the tab character and keyed by the fields
    :param str rna_bam: See reject_mutations:`rna_bam`
    :param int reject_threshold: See reject_mutations:`reject_threshold`
    :param float rna_min_alt_freq: See reject_mutations:`rna_min_alt_freq`
    :return: A named tuple of the True/False if the snv should be rejected or not, and a reason
    for rejection
    :rtype: reject_decision
    """
    if rna_bam is None:
        logging.warning('REJECTING INDEL %s:%s>%s.', indel['POS'] + 1, indel['REF'], indel['ALT'])
        return reject_decision(reject=True, reason='NoGenomeFile', coverage='.,./.,.', vaf=0.0)
    else:
        output_counts, reads = get_indel_alignment_info(rna_bam, indel)
        reject = None
        coverage = '/'.join([','.join([str(reads[x]), '.']) for x in ('covering', 'spanning')])
        low_rna_coverage = False
        if output_counts['ALT'] == 0:
            if reads['covering'] < reject_threshold and reads['spanning'] < reject_threshold:
                logging.debug('Mutation at position %s:%s has no evidence of existence in the '
                              'RNA-seq bam. However the coverage at the region (%s/%s '
                              'reads::Covering/spanning) is below the threshold for rejecting '
                              '(%s). Accepting.', indel['CHROM'], indel['POS'] + 1,
                              reads['covering'], reads['spanning'], reject_threshold)
                reject = reject_decision(reject=False, reason='LowRNACoverage', coverage=coverage,
                                         vaf=0.0)
            else:
                logging.warning('Mutation at position %s:%s has no evidence of existence in the '
                                'RNA-seq bam. Coverage = %s/%s reads (Covering/spanning). '
                                'Rejecting.', indel['CHROM'], indel['POS'] + 1, reads['covering'],
                                reads['spanning'])
                reject = reject_decision(reject=True, reason='NoAltDetected', coverage=coverage,
                                         vaf=0.0)
        else:
            # There was some evidence of the indel
            alt_freq = 1.0 * output_counts['ALT'] / sum([output_counts[x] for x in output_counts])
            if alt_freq < rna_min_alt_freq:
                logging.warning('Mutation at position %s:%s has less than %s ALT allele '
                                'frequency (%s) in the RNA-seq bam. Rejecting.',
                                indel['CHROM'], indel['POS'] + 1, rna_min_alt_freq,
                                round(alt_freq, 2))
                reject = reject_decision(reject=True, reason='LowAltFrequency', coverage=coverage,
                                         vaf=0.0)
            else:
                logging.debug('Accepted mutation at position %s:%s with %s read coverage and %s '
                              'VAF.',
                              indel['CHROM'], indel['POS'] + 1, coverage, round(alt_freq, 2))
                reject = reject_decision(reject=False, reason='.', coverage=coverage, vaf=alt_freq)
        return reject


def get_indel_alignment_info(rna_bam, vcf_record):
    """
    Get information about the spanning and covering reads at a given genomic locus for an indel.

    :param str rna_bam: Path to the RNA-Seq bam file
    :param dict vcf_record: A single vcf record split by the tab character and keyed by the fields
    :return: see get_alignment_info:`return`
    :rtype: tuple(collections.Counter|collections.defaultdict(Collections.Counter), dict(dict))
    """

    indel_length = len(vcf_record['REF']) - len(vcf_record['ALT'])
    deletion = indel_length > 0  # If deletion is false, it has to be an insertion
    indel_length = abs(indel_length)

    samfile = pysam.Samfile(rna_bam, 'rb')

    reads = {'spanning': 0, 'covering': 0}
    output_counts = {'REF': 0, 'ALT': 0}

    pileups = samfile.pileup(vcf_record['CHROM'], vcf_record['POS'],
                             vcf_record['POS'] + 1, truncate=True)
    for pileup in pileups:
        for read in pileup.pileups:
            reads['spanning'] += 1
            if read.is_del and read.is_refskip:
                # This is a spliced alignment. We will not consider it
                continue
            reads['covering'] += 1
            if read.query_position is None:
                # This is possibly a non-left-aligned deletion (maybe with a mutation)
                # REF:  A C T G A C T G A C T G A C T G A C T G A C T G
                # Rd1:                  A C T G A C T G-----T G A C T G
                # Rd2:                  A C T G A C T-----T T G A C T G    << mismatch in T
                #
                # read.indel will be 0 since this is inside an indel.
                pass
            elif read.indel == 0:
                # No indel on this read
                output_counts['REF'] += 1
                continue
            elif read.alignment.seq[read.query_position] != vcf_record['REF'][0]:
                logging.warning('read %s has %s at pos %s. expected %s', read.alignment.query_name,
                                read.alignment.seq[read.query_position], vcf_record['POS'] + 1,
                                vcf_record['REF'][0])

            if deletion:
                if read.indel == -indel_length:
                    if (ord(read.alignment.qual[read.query_position]) >= 63 and
                                ord(read.alignment.qual[read.query_position+1]) >= 63):
                        output_counts['ALT'] += 1
                else:
                    # TODO figure out left indel stuff
                    # Not a deletion of the required length (we will handle non-left-aligned
                    # events in the future.
                    continue
            else:  # Insertion
                if read.indel == indel_length:
                    quals = [ord(read.alignment.qual[read.query_position + x])
                             for x in range(0, indel_length+1)]
                    if sum(quals)/(indel_length+1) >= 63:
                        # On average, the bases must be Q30 or higher
                        output_counts['ALT'] += 1
                else:
                    # TODO figure out left indel stuff
                    # Not an insertion of the required length (we will handle non-left-aligned
                    # events in the future.
                    continue
        break  # Use only the first pileup position
    samfile.close()
    return output_counts, reads


def get_exon_start_pos(transcript_name, peptide_pos, exons, cds_starts):
    """
    Get the start position (1-based) of the exon containing `peptide_pos` in `peptide_name`.

    :param str transcript_name: The peptide name
    :param int peptide_pos: A protein position in a peptide
    :param dict exons: See return value of `transgene::get_exons`
    :param dict cds_starts: See return value of `transgene::get_exons`
    :return: The start position and a value identifying if the gene is on the positive strand
    :rtype: int, bool
    """
    if transcript_name not in exons:
        logging.warning('Transcript %s was not present in the GTF.', transcript_name)
        return None, None
    if transcript_name not in cds_starts:
        logging.warning('Transcript %s does not have a registered CDS start.', transcript_name)
        return None, None

    positive_strand = exons[transcript_name][0].strand == '+'
    if positive_strand:
        transcript_exons = [(exon.start, exon.end) for exon in exons[transcript_name]
                            if exon.end >= cds_starts[transcript_name]]
        transcript_exons[0] = (cds_starts[transcript_name], transcript_exons[0][1])
    else:
        transcript_exons = [(exon.start, exon.end) for exon in exons[transcript_name]
                            if exon.start <= cds_starts[transcript_name]]
        transcript_exons[0] = (transcript_exons[0][0], cds_starts[transcript_name]+2)

    frame = 0  # identifies the starting frame for the current exon
    total = 0  # tracks the total number of AAs accounted for at the end of the prev exon
    for idx, exon in enumerate(transcript_exons):
        exon_len = frame + exon[1] - (exon[0] - 1)
        if total + exon_len/3 >= peptide_pos:
            # The codon lies in this exon
            diff = peptide_pos - total
            if diff == 1 and frame != 0:
                # This means the start is in the previous exon
                if positive_strand:
                    return transcript_exons[idx-1][1] - frame + 1, True
                else:
                    return transcript_exons[idx-1][0] + frame - 1, False
            else:
                if frame != 0:
                    diff -= 1  # codon starting previous exon
                if positive_strand:
                    return exon[0] + (diff - 1) * 3 + ((3 - frame) if frame else 0), True
                else:
                    return exon[1] - (diff - 1) * 3 - ((3 - frame) if frame else 0), False
        else:
            total += exon_len/3
            frame = exon_len % 3


def get_codon(transcript_name, peptide_pos, exons, cds_starts, chrom):
    """
    Get the codon for a given amino acid in a transcript

    :param str transcript_name: The peptide name
    :param int peptide_pos: A protein position in a peptide
    :param dict exons: See return value of `transgene::get_exons`
    :param dict cds_starts: See return value of `transgene::get_exons`
    :param str chrom: The sequence for the chromosome containing the transcript
    :return: The codon
    :rtype: str
    """
    codon_start, positive_strand = get_exon_start_pos(transcript_name, peptide_pos, exons,
                                                     cds_starts)
    if codon_start is None:
        # Could not find the transcript, or it didn't have a CDS start
        return None

    if positive_strand:
        for i, exon in enumerate(exons[transcript_name]):
            if exon.end - codon_start in (0, 1):
                # This is a spliced exon and needs to be handled properly
                lhs = exon.end - codon_start + 1  # Number of nucleotides on the left hand side
                next_exon_start = exons[transcript_name][i + 1].start
                codon = (chrom[codon_start - 1:codon_start - 1 + lhs] +
                         chrom[next_exon_start - 1:next_exon_start - 1 + (3 - lhs)])
                break
        else:
            codon = chrom[codon_start-1:codon_start+2]
    else:
        for i, exon in enumerate(exons[transcript_name]):
            if codon_start - exon.start in (0, 1):
                # This is a spliced exon and needs to be handled properly
                rhs = codon_start - exon.start + 1  # Number of nucleotides on the right hand side
                next_exon_end = exons[transcript_name][i + 1].end
                codon = (chrom[next_exon_end - (3 - rhs):next_exon_end] +
                         chrom[codon_start - rhs:codon_start])
                break
        else:
            codon = chrom[codon_start - 3:codon_start]
        codon = codon[::-1].translate(trans)

    return codon.upper()



from __future__ import print_function

import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand
from version import version

# Set up a test class
# noinspection PyAttributeOutsideInit
class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        # Sanitize command line arguments to avoid confusing Toil code attempting to parse them
        sys.argv[1:] = []
        err_number = pytest.main(self.pytest_args)
        sys.exit(err_number)


setup(name='transgene',
      version=version,
      package_dir={'': 'src'},
      packages=find_packages('src', exclude=['*.test']),
      description='Translation of genomic events to proteomic space',
      url='http://github.com/arkal/transgene',
      author='Arjun Arkal Rao',
      author_email='aarao@ucsc.edu',
      license='Apache',
      install_requires=[
          'pysam>=0.9.1',
          'swalign>=0.3.3',
      ],
      tests_require=[
          'pytest==2.8.3'],
      test_suite='transgene',
      entry_points={
          'console_scripts': [
              'transgene = transgene.core:run_transgene']},
      cmdclass={'test': PyTest},
      zip_safe=False)



from __future__ import print_function

import collections
import csv
import logging
import pickle
import re
from cStringIO import StringIO

import swalign
from transgene.common import read_fasta, trans, BEDPE, translate


def get_transcriptome_data(infile, drop_transcript_version=False):
    """
    Parses Gencode transcript FASTA file and returns CDS sequences keyed by the transcript ID

    :param file infile: Gencode FASTA file object
    :param bool drop_transcript_version: Drop the version part of transcript ids.
    :return: {transcript ID: transcripts} and {gene ID: transcript IDs}
    :rtype: tuple
    """
    regex = r"(?P<transcript_id>ENST[0-9A-Z]+.\d+)\|(?P<gene_id>ENSG[0-9A-Z]+.\d+)" \
            r".*CDS:(?P<start>\d+)-(?P<stop>\d+)"
    gene_transcripts = collections.defaultdict(list)
    transcript_cds = {}
    for header, comment, seq in read_fasta(infile, 'ACGT'):
        match = re.search(regex, header)
        if match:
            # Remove the version number on gene ID
            gene_id, version = match.group("gene_id").split('.')
            transcript_id = match.group("transcript_id")
            if drop_transcript_version:
                transcript_id = transcript_id.split('.')[0]
            # GTF is one-based. Make interval [zero, one-based)
            start = int(match.group("start")) - 1
            stop = int(match.group("stop"))
            cds = seq[start: stop]
            gene_transcripts[gene_id].append(transcript_id)
            # Save transcript to gene ID mapping
            gene_transcripts[transcript_id] = match.group("gene_id")
            transcript_cds[transcript_id] = cds
    return transcript_cds, gene_transcripts


def rna_gene_in_bedpe(record):
    """
    Determine if one of the two candidates in a BEDPE line is an rna gene.

    :param BEDPE record: A BEDPE line from the input file
    :returns: True if one of the candidates is an RNA gene and False if not
    :rtype: bool
    """
    #  We will accept fusions that have an RP11- (lncRNA) 3' partner since they can still be
    # translated. This is a heuristic.
    return 'RP11-' in record.hugo1


def readthrough_in_bedpe(record, annotation, rt_threshold):
    """
    Determine if the two genes in the record are within `rt_threshold` bp of each other on the same
    chromosome.

    :param BEDPE record: A BEDPE line from the input file
    :param dict(str, GTFRecord) annotation: see `read_fusions:gene_annotations`
    :param rt_threshold: The genomic distance on the same chromosome below which we will call a
           candidate fusion a readthrough.
    :returns: True if the pair is considered a readthrough and False if not
    :rtype: bool
    """
    return (record.chrom1 == record.chrom2 and
            ((annotation[record.hugo1].start <= annotation[record.hugo2].start <=
                annotation[record.hugo1].end + rt_threshold) or
             (annotation[record.hugo2].start <= annotation[record.hugo1].start <=
                annotation[record.hugo2].end + rt_threshold)))


def read_fusions(fusion_file, gene_annotations, filter_mt, filter_ig, filter_rg, filter_rt,
                 rt_threshold, out_bedpe):
    """
    Reads in gene fusion predictions in modified BEDPE format.
    In addition to the basic BEDPE features, this function requires the fusion
    junction sequences and HUGO names for the donor and acceptor genes.

    :param file fusion_file: Fusion calls in BEDPE format
    :param dict(str, GTFRecord) gene_annotations: The gene annotations from the gtf
    :param bool filter_mt: Filter mitochondrial events?
    :param bool filter_ig: Filter immunoglobulin pairs?
    :param bool filter_rg: Filter RNA-Gene events?
    :param bool filter_rt: Filter transcriptional read-throughs?
    :param int rt_threshold: Distance threshold to call a readthrough
    :param file out_bedpe: A file handle to an output BEDPE file
    :returns: list of BEDPE namedtuples
    :rtype: list

    Modified BEDPE format
    chrom1:         Chromosome of first feature
    start1:         Zero-based starting position of the first feature or '.'
    end1:           One-based ending position of the first feature -- 5' fusion breakpoint
    chrom2:         Chromosome of second feature
    start2:         Zero-based starting position of the second feature -- 3' fusion breakpoint
    end2:           One-based ending position of thh second feature or '.'
    name:           Hyphenated Ensembl gene IDs (i.e. ENSG00000168172-ENSG00000165731)
    score:          Optional fusion score
    strand1:        Strand of first feature
    strand2:        Strand of second feature
    junctionSeq1:   Fusion junction sequence in first feature
    junctionSeq2:   Fusion junction sequence in second feature
    hugo1:          HUGO name for first feature
    hugo2:          HUGO name for second feature
    """

    calls = []

    for line in csv.reader(fusion_file, delimiter='\t'):
        if line[0].startswith('#'):
            print('\t'.join(line), file=out_bedpe)
            continue
        try:
            record = BEDPE(*line)
        except TypeError:
            raise ValueError("ERROR: fusion file is malformed.\n{}".format(read_fusions.__doc__))

        if filter_mt and 'M' in record.chrom1 or 'M' in record.chrom2:
            logging.warning("Rejecting %s-%s for containing a Mitochondrial gene.", record.hugo1,
                            record.hugo2)
            continue
        elif filter_ig and record.hugo1.startswith('IG') and record.hugo2.startswith('IG'):
            # This will drop some Insulin-like growth factor (IGF) proteins but they have a lot of
            # homology too so its ok.
            logging.warning("Rejecting %s-%s an an Immunoglobulin gene pair.", record.hugo1,
                            record.hugo2)
            continue
        elif filter_rg and rna_gene_in_bedpe(record):
            logging.warning("Rejecting %s-%s for containing a 5' RNA gene.", record.hugo1,
                            record.hugo2)
            continue
        elif filter_rt and readthrough_in_bedpe(record, gene_annotations, rt_threshold):
            logging.warning("Rejecting %s-%s as a potential readthrough.", record.hugo1,
                            record.hugo2)
            continue
        else:
            logging.info("Accepting %s-%s for further study.", record.hugo1, record.hugo2)
            print('\t'.join(line), file=out_bedpe)
            calls.append(record)

    if not calls:
        logging.warning('Input bedpe file was empty or had no actionable mutations.')
    return calls

# Namedtuple for storing alignment metrics
# Needs to be global for pickling
AlignStats = collections.namedtuple('AlignStats',
                                    'qstart, qstop, rstart, rstop, insertions, deletions')


def align_filter(ref, query, mode, mismatches_per_kb=1):
    """
    Aligns query to reference CDS sequence using the Smith-Waterman algorithm.
    Returns None if the alignment is clipped at the fusion boundary.

    :param str ref: In-frame reference transcript
    :param str query: Query transcript
    :param str mode: 'donor' or 'acceptor'
    :param int mismatches_per_kb: Allowed number of mismatches per kilobase
                                  of the alignment
    :return: Alignment features
    :rtype: namedtuple
    """

    bound_regex_str = r'Query\s*:\s*(?P<qstart>\d*)\s*[\w-]*\s*(?P<qstop>\d*)\s*[\|\s]*\s*' \
                      r'Ref\s*:\s*(?P<rstart>\d*)\s*[\w-]*\s*(?P<rstop>\d*)'
    bounds_regex = re.compile(bound_regex_str)
    mismatch_regex = re.compile(r'Mismatches: (?P<mismatches>\d+)')

    # Use default swalign parameters
    match = 2
    mismatch = -1
    scoring = swalign.NucleotideScoringMatrix(match, mismatch)
    sw = swalign.LocalAlignment(scoring)
    alignment = sw.align(ref, query)

    # Count the number of insertions and deletions
    insertions = 0
    deletions = 0
    for chr, num in alignment.cigar:
        if chr == 'I':
            insertions += num

        elif chr == 'D':
            deletions += num

    # Next grab the alignment statistics
    sw_output = StringIO()
    alignment.dump(out=sw_output)
    dump = sw_output.getvalue()
    sw_output.close()

    num_mismatches = None
    m = mismatch_regex.search(dump)
    if m:
        num_mismatches = int(m.group('mismatches'))

    s = bounds_regex.search(dump)
    if s:
        qstart = int(s.group('qstart')) - 1    # Make zero-based
        qstop = int(s.group('qstop'))

        # Filter alignments that have more than the allowed number of mismatches per kilobase
        if num_mismatches > int( mismatches_per_kb * (qstop - qstart) ):
            logging.debug("Mismatch filter: %d > %d" % (num_mismatches,
                                                        int(mismatches_per_kb * (qstop - qstart))))
            return

        # Filter alignments that do not include the donor breakpoint
        if mode == 'donor' and qstop != len(query):
            logging.debug('Donor alignment does not reach fusion boundary')
            return

        # Filter alignments that do not include the acceptor breakpoint
        elif mode == 'acceptor' and qstart != 0:
            logging.debug('Acceptor alignment does not reach fusion boundary')
            return

        rstart = int(s.group('rstart')) - 1    # Make zero-based
        rstop = int(s.group('rstart'))

        return AlignStats(qstart, qstop, rstart, rstop, insertions, deletions)


def scan_frame(reference_start):
    """
    Find the frame of a sequencing using the alignment starting position

    :param int reference_start: Alignment start position
    :return: Number of bases to slice sequence to make it in-frame
    :rtype: int
    """
    in_frame_adjustment = 0
    while (reference_start + in_frame_adjustment) % 3 != 0:
        in_frame_adjustment += 1

        if in_frame_adjustment > 3:
            return

    return in_frame_adjustment


def get_donor_junction_exon(breakpoint, exons):
    """
    Finds the first exon before the fusion breakpoint

    :param str transcript_id: donor Ensembl transcript ID
    :param int breakpoint: Genomic position of donor breakpoint
    :param list exons: Sorted list of GTFRecord objects keyed
    :return: Donor junction sequence
    :rtype: str
    """

    # Donor
    # We want the first exon that is upstream of the breakpoint
    #            exon.start < breakpoint        exon.start < breakpoint < exon.end
    #            exon.end  <= breakpoint
    #          _________     |      ___________           ______|_____
    # -->--->--|___1____|----|-----|__________|-----------|_____|_____|--------
    #                        |                                  |
    #                                                                       Traverse <--

    if exons[0].strand == '+':
        for exon in exons[::-1]:
            if exon.start < breakpoint and exon.end <= breakpoint:
                return exon.sequence

            elif exon.start < breakpoint < exon.end:
                # Gencode and breakpoint annotations are one-based
                return exon.sequence[: breakpoint - exon.start]


    # Donor
    # Want first exon that is downstream of breakpoint
    # exon.start < breakpoint < exon.end                  exon.start >= breakpoint
    #                                                     exon.end > breakpoint
    #          ____|_____          ___________      |      ___________
    # --<---<--|___|____|---------|__________|------|-----|_____1_____|--------
    #              |                                |
    # Traverse -->

    elif exons[0].strand == '-':
        for exon in exons[::-1]:
            if exon.start >= breakpoint and exon.end > breakpoint:
                return exon.sequence.translate(trans)[::-1]

            elif exon.start < breakpoint < exon.end:
                seq = exon.sequence[breakpoint - exon.start:]
                return seq.translate(trans)[::-1]


def get_acceptor_junction(breakpoint, exons):
    """
    Finds the first exon after the fusion breakpoint

    :param str transcript_id: donor Ensembl transcript ID
    :param int breakpoint: Genomic position of donor breakpoint
    :param dict exons:
    :return:
    """

    # Acceptor
    # Want first exon downstream of breakpoint
    #               exon.start >= breakpoint      exon.start < breakpoint < breakpoint
    #               exon.end  > breakpoint
    #          _________     |      ___________           ______|_____
    # -->--->--|_______ |----|-----|_____x_____|----------|_____|_____|--------
    #                        |                                  |
    # Traverse -->

    if exons[0].strand == '+':
        for exon in exons:
            if exon.start >= breakpoint and exon.end > breakpoint:
                return exon.sequence

            elif exon.start < breakpoint < exon.end:
                seq = exon.sequence[breakpoint - exon.start:]
                return seq.translate(trans)[::-1]

    # Acceptor
    # Want first exon downstream of breakpoint
    #               exon.start >= breakpoint      exon.start < breakpoint < breakpoint
    #               exon.end  > breakpoint
    #          _________     |      ___________           ______|_____
    # --<---<--|_______ |----|-----|_____x_____|----------|_____|_____|--------
    #                        |                                  |
    #                                                                   <-- Traverse

    elif exons[0].strand == '-':
        for index, exon in enumerate(exons):
            if exon.start < breakpoint and exon.end <= breakpoint:
                return exon.sequence.translate(trans)[::-1]

            elif exon.start < breakpoint < exon.end:
                seq = exon.sequence[: breakpoint - exon.start]
                return seq.translate(trans)[::-1]


def insert_fusions(transcriptome, fusion_calls, gene_transcripts, peplen, outfile, exons=None, output_frameshift=False):
    """
    Generates fusion peptides for MHC binding prediction. If the junction sequence is not
    provided for the donor and acceptor, a theoretical sequence is inferred using a
    genome annotation file.

    :param dict transcriptome: Dictionary of transcripts keyed by transcript ID
    :param list[BEDPE] fusion_calls: List of BEDPE namedtuples
    :param dict[list] gene_transcripts: Dictionary of transcript IDs keyed by gene ID
    :param int peplen: Length of n-mer peptide
    :param file outfile: File object to write n-mer peptides in FASTA format
    :param bool output_frameshift: Output peptides that result from a frameshift in the acceptor sequence
    :param dict exons: Dictionary mapping transcript IDs to ordered list of GTFRecord objects
    """
    logging.info('Generating fusion peptides of length %d' % peplen)

    found = False

    # Load previous alignments
    try:
        with open('transgene_fusion_alignments.pkl') as f:
            previous_alignments = pickle.load(f)

    except (OSError, IOError):
        previous_alignments = {}

    # Iterate over fusion calls
    for call in fusion_calls:

        logging.info('Processing fusion: %s' % repr(call.name))

        donor_name, acceptor_name = call.name.split('-')

        junction_seq1 = call.junctionSeq1
        end1 = int(call.end1)

        # Iterate over all possible transcripts for the donor gene
        for donor_transcript_id in gene_transcripts[donor_name]:

            # If a donor sequence is not provided, find the theoretical one
            if exons and exons[donor_transcript_id] and call.junctionSeq1 == '.':
                logging.debug("Inferring donor junction sequence using breakpoint")
                junction_seq1 = get_donor_junction_exon(end1, exons[donor_transcript_id])

            if junction_seq1 is None:
                continue

            # Check to see if the junction sequence is large enough for neoepitope prediction
            if len(junction_seq1) // 3 < peplen - 1 :
                logging.warn('Donor sequence is not long enough for neoepitope prediction'
                             '\n%s' % junction_seq1)
                continue

            # Find the frame in the donor sequence using a reference transcriptome
            try:
                donor_transcript = transcriptome[donor_transcript_id]

                # Check if we have already aligned this sequence (ie for another peptide length)
                try:
                    bounds = previous_alignments[(donor_transcript, junction_seq1)]

                # If not, then align the sequence and save the alignment information
                except KeyError:
                    bounds = align_filter(donor_transcript, junction_seq1, 'donor')
                    previous_alignments[(donor_transcript, junction_seq1)] = bounds

                # Donor sequence may not align to some isoforms
                if bounds is None:
                    logging.debug('Donor sequence did not align to reference transcripts')
                    continue

            except KeyError:
                continue

            # Find the position of the transcript where the predicted donor sequence starts
            in_frame_adjustment = scan_frame(bounds.rstart)
            if in_frame_adjustment is None:
                continue

            # Slice the donor transcript such that the first base is in-frame
            # May have an overhang base
            donor_start = bounds.qstart + in_frame_adjustment
            in_frame_donor_seq = junction_seq1[donor_start:]
            in_frame_donor_seq_len = len(in_frame_donor_seq)

            if acceptor_name in gene_transcripts:
                acceptor_transcript_ids = gene_transcripts[acceptor_name]

            else:
                logging.debug('Could not find acceptor transcript ID')
                break

            # Now iterate over all possible acceptor transcript sequences
            start2 = int(call.start2)
            for acceptor_transcript_id in acceptor_transcript_ids:
                junction_seq2 = call.junctionSeq2

                # Predict acceptor sequence if none was provided
                if exons and acceptor_transcript_id in exons and junction_seq2 == '.':
                    junction_seq2 = get_acceptor_junction(start2, exons[acceptor_transcript_id])

                if junction_seq2 is None:
                    logging.debug('Could not infer theoretical acceptor junction sequence')
                    continue

                acceptor_in_frame_adjustment = None
                try:
                    # Get reference coding transcript for acceptor gene
                    acceptor_transcript = transcriptome[acceptor_transcript_id]

                    try:
                        # Find the position of the transcript where the predicted acceptor sequence starts
                        acceptor_bounds = previous_alignments[(acceptor_transcript, junction_seq2)]

                    except KeyError:
                        acceptor_bounds = align_filter(acceptor_transcript, junction_seq2, 'donor')
                        previous_alignments[(acceptor_transcript, junction_seq2)] = bounds

                    # If the acceptor sequence could not be found, move onto the next transcript
                    if acceptor_bounds is None:
                        continue

                    acceptor_in_frame_adjustment = scan_frame(acceptor_bounds.rstart)
                    if acceptor_in_frame_adjustment is None:
                        continue

                # We can still generate a fusion transcript without a reference for the acceptor gene
                except KeyError:
                    pass

                # Figure out if frame-shift occurred as a result of the fusion
                acceptor_frame_shift = False

                # If we don't have a reference transcript, then just continue without this information
                if acceptor_in_frame_adjustment is None:
                    pass

                # There could be an indel that corrects the frameshift downstream
                elif (in_frame_donor_seq_len % 3 + acceptor_in_frame_adjustment) % 3 != 0:
                    logging.warn('%s--%s has a 3\' frame shift' % (call.hugo1, call.hugo2))
                    acceptor_frame_shift = True

                # Use the donor and acceptor sequences to generate the fusion transcript
                fusion_seq = in_frame_donor_seq + junction_seq2
                logging.debug(fusion_seq)

                # The fusion peptide for MHC binding prediction includes all sliding windows that would
                # generate a novel peptide sequence. Therefore, the donor sequence starts at the position
                # of the last donor codon minus the fusion peptide length plus two codons. The acceptor
                # portion consists of the last donor codon plus the peptide length.
                #
                #  Transcript Start: 0  3  6  9  12 15
                #    Transcript End: 2  5  8  11 14 17
                #       Codon Index: 0  1  2  3  4  5
                #    Fusion Protein: D  D  D  A  A  A
                #
                #                       D  D  A
                #                          D  A  A
                #              Result:  D  D  A  A
                #
                #   peplen = 9
                #
                #   start = 6 - 9 + 6 = 3
                #   end   = 6 + 9     = 14

                last_donor_codon = in_frame_donor_seq_len - in_frame_donor_seq_len % 3 - 3

                start = last_donor_codon - 3 * peplen + 6

                end = last_donor_codon + 3 * peplen

                in_frame_junction_seq = fusion_seq[start: end]

                fusion_peptide = translate(in_frame_junction_seq)
                logging.debug("Fusion peptide:\n%s" % fusion_peptide)

                if '*' in fusion_peptide:
                    logging.warn('Skipping fusion epitope containing stop codon')
                    continue


                # This may occur if the fusion sequence occurs near the end of the transcript. Warn the user, but
                # still output the fusion peptide because there is at least one fusion sequence that can be generated
                # from it.
                if len(fusion_peptide) < peplen:
                    logging.warning('Fusion peptide %s-%s is malformed.\n'
                                    'Expected peptide of length %d, got %d' % (donor_name,
                                                                               acceptor_name,
                                                                               2 * peplen - 2,
                                                                               len(fusion_peptide)))
                    break

                gencode_donor_id = gene_transcripts[donor_transcript_id]
                gencode_acceptor_id = gene_transcripts[acceptor_transcript_id]

                # Write fusion neoepitope sequence to FASTA file
                fasta = '>{donor_gene}-{acceptor_gene}_' \
                        '{donor_transcript}-{acceptor_transcript}_' \
                        '{donor_hugo}-{acceptor_hugo}_' \
                        'FUSION_{score}\n' \
                        '{sequence}\n'.format(donor_gene=gencode_donor_id,
                                              acceptor_gene=gencode_acceptor_id,
                                              donor_transcript=donor_transcript_id,
                                              acceptor_transcript=acceptor_transcript_id,
                                              donor_hugo=call.hugo1,
                                              acceptor_hugo=call.hugo2,
                                              score=call.score,
                                              sequence=fusion_peptide)
                outfile.write(fasta)
                found = True

                # A frameshift in the acceptor sequence also generates
                # candidate neoepitope sequences
                if output_frameshift and acceptor_frame_shift:
                    for i in range(end, len(fusion_seq) - 3 * peplen, 3):
                        shift_sequence = fusion_seq[i: i + 3 * peplen]
                        shift_peptide = translate(shift_sequence)

                        if '*' in shift_peptide:
                            break

                        if len(shift_peptide) != peplen:
                            break

                        logging.debug('Created fusion frameshift peptide:\n%s' % shift_peptide)


                        # Write frameshift peptide to FASTA file
                        fasta = '>{donor_gene}-{acceptor_gene}_' \
                                '{donor_transcript}-{acceptor_transcript}_' \
                                '{donor_hugo}-{acceptor_hugo}_' \
                                'FUSION_ACCEPTOR_FRAMESHIFT\n' \
                                '{sequence}\n'.format(donor_gene=gencode_donor_id,
                                                      acceptor_gene=gencode_acceptor_id,
                                                      donor_transcript=donor_transcript_id,
                                                      acceptor_transcript=acceptor_transcript_id,
                                                      donor_hugo=call.hugo1,
                                                      acceptor_hugo=call.hugo2,
                                                      sequence=shift_peptide)
                        outfile.write(fasta)

    if not found:
        logging.warn("Did not generate any fusion peptides from data!")

    # Save alignments for different peptide lengths
    with open('transgene_fusion_alignments.pkl', 'w') as f:
        pickle.dump(previous_alignments, f)




#!/usr/bin/env python2.7
"""
Author : Arjun Arkal Rao
Affiliation : UCSC BME, UCSC Genomics Institute
File : Transgene/transgene


Program info can be found in the docstring of the main function
"""
from __future__ import division, print_function

import argparse
import collections
import json
import logging
import os
import random
import shutil
import string
import sys
import time
from copy import deepcopy
from functools import partial
from multiprocessing import Manager, Pool
from operator import itemgetter
from tempfile import mkstemp

import pysam
from transgene.common import (chrom_sort,
                              first_mismatch,
                              file_type,
                              genetic_code,
                              get_exons,
                              get_snpeff_3_6_changes,
                              get_snpeff_4_0_changes,
                              get_snpeff_4_x_changes,
                              get_vep_changes,
                              read_fasta,
                              read_genes_from_gtf,
                              reject_decision,
                              trans,
                              translate)
from transgene.fusion import get_transcriptome_data, insert_fusions, read_fusions
from transgene.indel import get_codon, get_exon_start_pos, reject_indel
from transgene.snv import reject_snv


def reject_mutation(call, rna_bam=None, reject_threshold=None, rna_min_alt_freq=None, dna_bam=None,
                    oxog_min_alt_freq=None, reject_indels=False):
    """
    Decide whether the mutation should be rejected based on the expression filter.

    :param dict call: A single vcf record split by the tab character and keyed by the fields
    :param str rna_bam: The path to the RNA-seq file. Can be None suggesting that expression
           filtering should not be carried out.
    :param int reject_threshold: The read coverage above which we reject a no-ALT-coverage event
    :param float rna_min_alt_freq: The ALT allele frequency below which calls will be rejected as
           being having insufficient RNA evidence.
    :param str dna_bam:  The path to the tumor DNA-seq file. Can be None suggesting that oxoG
           filtering should not be carried out.
    :param float oxog_min_alt_freq: The ALT allele frequency below which calls will be rejected as
           being oxoG artifacts.
    :param bool reject_indels: Should we reject indels?
    :return: A named tuple of the True/False if the mutation should be rejected or not, and a reason
    for rejection
    :rtype: tuple(reject_decison)
    """
    # We need either a rna bam or a dna bam to operate
    assert rna_bam or dna_bam, "Cannot filter without at least one bam file."
    logging.info('Processing %s>%s mutation at position %s:%s', call['REF'], call['ALT'],
                 call['CHROM'], call['POS'] + 1)
    if len(call['REF']) == len(call['ALT']):
        return reject_snv(call, rna_bam, reject_threshold, rna_min_alt_freq, dna_bam,
                          oxog_min_alt_freq),
    else:
        if ',' in call['ALT']:
            # MAV.
            alts = call['ALT'].split(',')
            decisions = []
            for i, alt in enumerate(alts):
                # Modify this for each call to reject_snv
                call['ALT'] = alt
                decisions.append(reject_snv(call, rna_bam, reject_threshold, rna_min_alt_freq,
                                            dna_bam, oxog_min_alt_freq))
            # Return it to the original state
            call['ALT'] = ','.join(alts)
            return tuple(decisions)
        else:
            # Indel
            if not reject_indels:
                return reject_indel(call, rna_bam, reject_threshold, rna_min_alt_freq),
            else:
                logging.warning('REJECTING INDEL %s:%s>%s.', call['POS'] + 1, call['REF'],
                                call['ALT'])
                return reject_decision(reject=True, reason='NoRNAFiltering', coverage='.,./.,.',
                                       vaf=0.0),


def write_to_vcf(data, out_vcf, header=False):
    """
    Write results to a vcf file
    :param dict data: The data line to write to the vcf
    :param file out_vcf: An open file handle to the output vcf
    :param header: Is this the field header
    :return: None
    """
    assert out_vcf.mode == 'w'
    # Hard coding the output order
    fields = ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO']
    if header:
        print('#', end='', sep='', file=out_vcf)
        print('\t'.join(fields), file=out_vcf)
    else:
        print('\t'.join([str(data[field]) if field in data else '.' for field in fields]),
              file=out_vcf)


def read_annotated_vcf(vcf_file, rna_file=None, out_vcf=None, reject_threshold=None,
                       rna_min_alt_freq=None, dna_file=None, oxog_min_alt_freq=None, processes=1,
                       reject_indels=False):
    """
    This module reads in the SNVs from the SnpEffed vcf file. It assumes that the SnpEff results
    have been writen to the INFO column. If the header contains a line that starts with  `#CHROM`,
    we will attempt to dynamically find the index of the `CHROM`, `POS`, `REF`, `ALT` and `INFO`
    columns.  If not, indexes 0, 1, 3, 4 and 7 (columns 1, 2, 4, 5 and 8) will be used.
    Example vcf column header:
    #CHROM  POS     ID      REF     ALT     QUAL    FILTER  INFO

    :param file vcf_file: An open file handle to the snpeff/vep vcf file.
    :param str rna_file: The path to the RNA-seq file. Can be None suggesting that expression
           filtering should not be carried out.
    :param file out_vcf: An open file handle to the output vcf file
    :param int reject_threshold: The read coverage above which we reject a no-ALT-coverage event
    :param float rna_min_alt_freq: The ALT allele frequency below which calls will be rejected as
           being having insufficient RNA evidence.
    :param str dna_file:  The path to the tumor DNA-seq file. Can be None suggesting that oxoG
           filtering should not be carried out.
    :param float oxog_min_alt_freq: The ALT allele threshold below which calls will be rejected as
           being oxoG artifacts.
    :param int processes: How many processes should handle the parsing?
    :param bool reject_indels: Should we reject indels?
    :returns: A parsed dictionary each of snvs and indels to be processed, and the annotator
    :rtype: tuple(dict, str)
    """
    if out_vcf:
        assert out_vcf.mode == 'w', 'Output vcf file is not open for writing'
    logging.info('Reading in Mutations from %s', vcf_file.name)
    indexes = {'CHROM': 0,
               'POS': 1,
               'REF': 3,
               'ALT': 4,
               'INFO': 7}

    annotator = None
    # Set up the Manager, and the shared dict and queue
    manager = Manager()
    vcf_queue = manager.Queue()
    mutations = manager.dict()
    # Read in the vcf header and identify the columns in the file
    for line in vcf_file:
        if line.startswith('#'):
            if line.startswith('#CHROM'):
                # This is the last line of the header.
                line = line.strip().lstrip('#').split('\t')
                temp_indexes = {val: pos for pos, val in enumerate(line)}
                for column in indexes.keys():
                    if column in temp_indexes:
                        logging.debug('Dynamically identified %s column from vcf header to be '
                                      'column %s.', column, temp_indexes[column] + 1)
                        indexes[column] = temp_indexes[column]
                    else:
                        logging.warning('Could not dynamically identify %s column from vcf header. '
                                        'Using column %s.', column, indexes[column] + 1)
                if rna_file or dna_file:
                    print('##INFO=<ID=reason,Number=.,Type=String,Description=Reason for rejecting/'
                          'accepting a mutation.', file=out_vcf)
                    print('##INFO=<ID=coverage,Number=1,Type=String,Description=Reads supporting '
                          'the decision (Covering_RNA,Covering_DNA/Spanning_RNA,Spanning_DNA).',
                          file=out_vcf)
                    print('##INFO=<ID=vaf,Number=1,Type=String,Description=The >Q30 variant '
                          'allele frequency for the mutation in the RNA-Seq bam.', file=out_vcf)
                    write_to_vcf(indexes, out_vcf, header=True)
            elif rna_file or dna_file:
                print(line, end='', file=out_vcf)
            if 'VEP=' in line:
                annotator = 'VEP'
            elif 'SnpEffVersion' in line:
                annotator = 'SNPEff' + line.split()[0].split('"')[1].translate(None, string.letters)
            if annotator not in ('SNPEff3.6', 'vep') and 'SnpEffCmd' in line and 'classic' in line:
                # If we run SNPEff 4.0+ with the -classic flag
                annotator = 'SNPEff3.6'
            continue
        elif len(line.strip()) == 0:
            continue
        else:
            # Populate the queue.
            vcf_queue.put(line)
    for _ in range(0, processes):
        # Add the graceful worker shutdown signal to the queue for every worker.
        vcf_queue.put(None)
    if annotator is None:
        raise RuntimeError('Could not determine annotator type from the VCF header. Please check '
                           'the integrity of the header and retry.')
    # Start the workers
    logging.info('Starting %s workers for vcf parsing.' % processes)
    pool = Pool(processes=processes)
    parse_vcf_line_partial = partial(parse_vcf_line,
                                     queue=vcf_queue,
                                     indexes=indexes,
                                     out_calls=mutations,
                                     rna_file=rna_file,
                                     annotator=annotator,
                                     reject_threshold=reject_threshold,
                                     rna_min_alt_freq=rna_min_alt_freq,
                                     dna_file=dna_file,
                                     oxog_min_alt_freq=oxog_min_alt_freq,
                                     reject_indels=reject_indels)
    pool.map(parse_vcf_line_partial, range(0, processes))
    pool.close()
    pool.join()
    # Now that all the processes have completed, we need to process the worker specific vcfs
    worker_vcfs = {i: '.worker_%s.vcf' % i for i in range(0, processes)}
    if rna_file or dna_file:
        if processes == 1:
            # The queue is ordered and the lone process would have run the lines in order so we can
            # just copy the contents over as-is without spending compute on sorting, etc.
            with open(worker_vcfs[0]) as v_file:
                for line in v_file:
                    print(line, end='', file=out_vcf)
        else:
            chrom_lines = collections.defaultdict(dict)
            for vcf in worker_vcfs:
                with open(worker_vcfs[vcf]) as v_file:
                    for line in v_file:
                        line_ = line.strip().split()
                        chrom_lines[line_[indexes['CHROM']]][int(line_[indexes['POS']])] = line
            for chrom in chrom_sort(chrom_lines.keys()):
                for pos in sorted(chrom_lines[chrom].keys()):
                    print(chrom_lines[chrom][pos], end='', file=out_vcf)
    for vcf in worker_vcfs:
        # Delete the temp vcfs now that we're done with them.
        os.remove(worker_vcfs[vcf])

    # Merge the dict structure into a dict of dicts
    out_muts = collections.defaultdict(dict)
    non_syn_seen = False

    for transcript, hugo, mut, tlen in mutations.keys():
        out_muts[transcript][mut] = mutations[(transcript, hugo, mut, tlen)]
        if 'has_indel' in out_muts[transcript]:
            out_muts[transcript]['has_indel'] = (out_muts[transcript]['has_indel'] or
                                                 out_muts[transcript][mut]['indel'] is not None)
        else:
            out_muts[transcript]['has_indel'] = out_muts[transcript][mut]['indel'] is not None

        if (not non_syn_seen and
                (out_muts[transcript][mut]['indel'] or (mut[0] != mut[-1] and mut[-1] != '*'))):
            # If we haven't seen a non_synonymous mutation, and this is non synonymous, update the
            # flag
            non_syn_seen = True
        # probably overkill
        if 'len' in out_muts[transcript]:
            assert out_muts[transcript]['len'] == int(tlen)

        else:
            out_muts[transcript]['len'] = int(tlen)
            out_muts[transcript]['HUGO'] = hugo

    if not (out_muts and non_syn_seen):
        logging.warning('Input vcf file was empty or had no actionable mutations.')
        out_muts = {}  # Return an empty dictionary

    return dict(out_muts), annotator


def get_ref_pos_alt_aa(aa_change):
    """
    Determine the reference AA(s), the position in the protein, and the alternate AA(s) for a given
    mutation in the form A123B

    :param aa_change: A String describing a mutation in protein space
    :return: reference AA(s), position, Alternamte AA(s)
    :rtype: tuple(str, int, str)
    """
    # We expect the mutation to be 3-parts, non-numeric, numeric, non-numeric or NA
    expected_isdigit = False
    parts = ['']
    for char in aa_change:
        if char.isdigit() is expected_isdigit:
            parts[-1] += char
        else:
            parts.append(char)
            expected_isdigit = not expected_isdigit
    parts[1] = int(parts[1])
    if len(parts) == 2:
        parts.append('')
    return parts


def get_codon_nuc_change(codon_change):
    """
    For a given codon change E.g. aAt/aGt, return the changed base ([A,G] in the example).

    :param str codon_change: The codon change
    :return: The modified ref and alt
    :rtype: tuple

    >>> get_codon_nuc_change('aAt/aGt')
    ('A', 'G')
    >>> get_codon_nuc_change('agG/agT')
    ('G', 'T')
    >>> get_codon_nuc_change('p.A>G')
    ('A', 'G')
    >>> get_codon_nuc_change('agg/agg')
    Traceback (most recent call last):
    ..
    AssertionError
    """
    if '/' in codon_change:
        ref, alt = codon_change.split('/')
    else:
        assert '>' in codon_change
        ref, alt = codon_change.split('>')
        ref = ref[-1]  # p.A >> A
    for i in range(3):
        # range(3) is ok because this will only happen with MAV
        if ref[i] == alt[i]:
            continue
        else:
            return ref[i], alt[i]
    assert False


def parse_vcf_line(worker_id, queue, indexes, out_calls, annotator,
                   rna_file=None, reject_threshold=None, rna_min_alt_freq=None, dna_file=None,
                   oxog_min_alt_freq=None, reject_indels=False, ):
    """
    Parse one mutation-containing line in the input vcf.

    :param int worker_id: The id for this worker
    :param multiprocessing.Manager.Queue queue: A queue that will supply a line of the vcf as every
           item
    :param dict indexes: A dict indicating the fields in the vcf line.
    :param multiprocessing.Manager.dict out_calls: The shared output vcf that will contain the
           output calls.
    :param str annotator: The Annotator used to annotate the vcf (VEP, SNPEFF_<VERSION>)
    :param str rna_file: The path to the RNA-seq file. Can be None suggesting that expression
           filtering should not be carried out.
    :param int reject_threshold: The read coverage above which we reject a no-ALT-coverage event
    :param float rna_min_alt_freq: The ALT allele frequency below which calls will be rejected as
           being having insufficient RNA evidence.
    :param str dna_file:  The path to the tumor DNA-seq file. Can be None suggesting that oxoG
           filtering should not be carried out.
    :param float oxog_min_alt_freq: The ALT allele threshold below which calls will be rejected as
           being oxoG artifacts.
    :param bool reject_indels: Should we reject indels?
    """
    logging.info('VCF parsing worker %s is up and running.' % worker_id)
    get_snpeff_changes = {'SNPEff3.6': get_snpeff_3_6_changes,
                          'SNPEff4.0': get_snpeff_4_0_changes,
                          'SNPEff4.1': get_snpeff_4_x_changes,
                          'SNPEff4.2': get_snpeff_4_x_changes,
                          'SNPEff4.3': get_snpeff_4_x_changes,
                          }
    # Unfortunately, this logic has to be here and cannot be abstracted into the individual
    # tool/version methods because we need to know if we can reject this line at this point for not
    # being annotated (intergenic variant)
    annot_field_sep = {'SNPEff3.6': 'EFF',
                       'SNPEff4.0': 'EFF',
                       'SNPEff4.1': 'ANN',
                       'SNPEff4.2': 'ANN',
                       'SNPEff4.3': 'ANN',
                       'VEP': 'CSQ'
                       }
    with open('.worker_%s.vcf' % worker_id, 'w') as out_vcf:
        while True:
            line = queue.get(timeout=30)
            if line is None:
                break
            line = line.strip().split('\t')
            line = {x: line[y] for x, y in indexes.items()}

            eff = [x for x in line['INFO'].split(';') if x.startswith(annot_field_sep[annotator])]
            alt_alleles = line['ALT'].split(',')
            if rna_file or dna_file:
                # NOTE: vcf snvs are 1-based and not 0-based!
                line['POS'] = int(line['POS']) - 1
                if eff:
                    decisions = reject_mutation(line, rna_file, reject_threshold, rna_min_alt_freq,
                                                dna_file, oxog_min_alt_freq, reject_indels)
                else:
                    logging.warning('Mutation at position %s:%s corresponds to a non-exonic '
                                    'region. Rejecting.', line['CHROM'], line['POS'] + 1)
                    decisions = (reject_decision(reject=True, reason='NonExonic', coverage='NA',
                                                 vaf='0.0'),)
                # Print the line to the output vcf
                line['INFO'] = ';'.join([line['INFO'],
                                         'reason=' + ','.join(d.reason for d in decisions),
                                         'coverage=' + str(decisions[0].coverage),
                                         'VAF=' + ','.join(str(d.vaf) for d in decisions)])
                line['FILTER'] = 'REJECT' if all(d.reject for d in decisions) else 'PASS'
                # Temporarily change back to 1-based for printing the vcf record
                line['POS'] += 1
                write_to_vcf(line, out_vcf)
                line['POS'] -= 1
                # If all ALT alleles were rejected, we can skip to the next call
                if all(d.reject for d in decisions):
                    continue
            else:
                if len(line['REF']) == len(line['ALT']) or len(alt_alleles) != 1:
                    # Accept snvs only
                    decisions = tuple([reject_decision(False, None, None, 0.0)
                                       for _ in alt_alleles])
                else:
                    logging.warning('Rejecting INDEL %s:%s>%s.', line['POS'], line['REF'],
                                    line['ALT'])
                    decisions = tuple([reject_decision(True, 'NoRNAFiltering', 'NA', 'NA')
                                       for _ in alt_alleles])

            if annotator == 'VEP':
                eff = get_vep_changes(eff[0])
            else:
                eff = get_snpeff_changes[annotator](eff[0])

            indel = True if len(line['REF']) != len(line['ALT']) and len(decisions) == 1 else None
            for di, decision in enumerate(decisions):
                if decision.reject:
                    # happens if one of the ALT alleles was rejected but the other wasn't
                    continue
                for i in range(0, len(eff)):
                    if not indel:
                        # MAVs will have a mix of both ALTs in the EFF line so we need to ensure
                        # this change corresponds to the current ALT
                        temp_codon_change = get_codon_nuc_change(eff[i].codon_change)
                        if temp_codon_change not in [(line['REF'], alt_alleles[di]),
                                                      (line['REF'].translate(trans),
                                                       alt_alleles[di].translate(trans))]:
                            continue
                    ref_aa, pos, alt_aa = get_ref_pos_alt_aa(eff[i].aa_change)
                    if not alt_aa or alt_aa.endswith('?'):
                        if indel:
                            # Only possible case for alt_aa to be '?'
                            if 'frameshift_variant' in eff[i].annotation:
                                if len(line['REF']) > len(line['ALT']):
                                    # Frame shift arising from a deletion. Either a full in-frame
                                    # codon deletion, or a frame shift
                                    indel = 'frame_shift_deletion'
                                else:
                                    indel = 'frame_shift_insertion'
                            else:
                                indel = 'full_codon_deleltion'
                        else:
                            assert not alt_aa.endswith('?'), '? seen in an SNV for the ALT AA'
                            if ref_aa == '?':
                                logging.warning('Cannot handle mutation (%s:%s) in gencode protein '
                                                '(%s) sequence. Annotator warnings were (%s)',
                                                line['CHROM'], line['POS'], eff[i].transcript_id,
                                                eff[i].warnings)
                                continue
                            else:
                                # Synonymous change
                                alt_aa = ref_aa
                    # At this point, indel is False, frame_shift_X, full_codon_deletion, or True.
                    # True implies it is a full codon insertion or a codon insertion with codon
                    # change.
                    if indel is True:
                        indel = 'full_codon_insertion'

                    out_calls[(eff[i].transcript_id, eff[i].gene_name, eff[i].aa_change,
                               eff[i].transcript_len)] = {
                        'AA': {'REF': ref_aa,
                               'ALT': alt_aa,
                               'change': eff[i].codon_change,
                               'VAF': str(round(decision.vaf, 2)),
                               'POS': pos},
                        'NUC': {'REF': line['REF'],
                                'ALT': alt_alleles[di],
                                'POS': line['POS'],
                                'CHROM': line['CHROM']},
                        'indel': indel}
    logging.info('VCF parsing worker %s received signal to go down.' % worker_id)


def build_groups(mut_list, base_mut=None, out_list=None):
    """
    Recursively propagates a binary tree of muts

    :param list base_mut: The base for the mutation
    :param list mut_list: The remaining mutations to add to the tree
    :param list out_list: The list to append values to
    :return: a list of lists containing all the possible mutations

    >>> build_groups(['a', 'b'])
    [('a', 'b'), ('a',), ('b',)]
    >>> build_groups(['a'], ['b'])
    [('b', 'a'), ('b',)]
    """
    # Since we can't make mutables the default parameters
    if base_mut is None:
        base_mut = []
    if out_list is None:
        out_list = []
    if not mut_list:
        # base case
        if base_mut:
            out_list.append(tuple(base_mut))
        return None
    else:
        new_base = [mut_list[0]]
        build_groups(mut_list[1:], base_mut + new_base, out_list)
        build_groups(mut_list[1:], base_mut, out_list)
        return out_list


def fix_group(grp, phased_groups):
    """
    This is used to fix a group that has a 0 in it, using the other groups
    :param tuple grp: The group to be fixed
    :param set phased_groups: a set of tuples of other groups
    :return:
    """
    index = grp.index(0)
    left_index = index - 1 if index > 0 else None
    right_index = index + 1 if index < len(grp)-1 else None
    possible_values = set([x[index]
                           for x in phased_groups
                           if ((left_index is not None and
                                x[left_index] == grp[left_index]) or
                               left_index is None) and
                           ((right_index is not None and
                             x[right_index] == grp[right_index]) or
                            right_index is None)])
    if not possible_values:
        logging.debug('Could not fix a group.', str(grp))
        return None
    else:
        grp = list(grp)
        for val in possible_values:
            if tuple(grp[:index] + [val] + grp[index+1:]) in phased_groups:
                return None
        # If we reach here, we haven't found a perfect match in phased_groups so we create one using
        # one of the possible values for the mutation.
        grp = tuple(grp[:index] + [random.choice(list(possible_values))] + grp[index+1:])
    if 0 in grp:
        return fix_group(grp, phased_groups)
    else:
        return grp


def get_mutation_groups(p_snvs, peplen, pfasta_name, rna_bam=None):
    """
    Produces a list of mutation groups for the given peptide

    :param dict p_snvs: parsed snvs from the input file
    :param int peplen: The length of the peptide
    :param str pfasta_name: The full name of the protein fasta (for debugging purposes)
    :param str rna_bam: path to the bam file
    :return: mutation groups
    :rtype: dict
    """
    mutations = [m for m in p_snvs if m != 'has_indel']
    mutations = sorted(mutations, key=lambda m: p_snvs[m]['AA']['POS'])
    groups = []
    while True:
        group_start = mutations[0]
        groups.append([group_start] + [y for y in mutations[1:]
                                       if p_snvs[group_start]['AA']['POS'] + peplen >
                                       p_snvs[y]['AA']['POS'] >=
                                       p_snvs[group_start]['AA']['POS']])
        for i in groups[-1][1:]:
            mutations.remove(i)
        mutations.remove(group_start)
        if not mutations:
            break
    mutation_groups = [build_groups(x) for x in groups]
    if rna_bam is None:
        logging.debug('Peptide %s detected to have the following mutation groups %s', pfasta_name,
                      mutation_groups)
        return mutation_groups
    else:
        out_groups = []
        for group in mutation_groups:
            # We will parse for co-expressed mutations
            if len(group) == 1:
                # This is guaranteed to have coverage because we have already done a coverage filter
                out_groups.append([list(group[0])])  # List to make it consistent with bigger groups
                continue
            samfile = pysam.AlignmentFile(rna_bam, 'rb')
            mutations = group[0]
            if p_snvs[mutations[0]]['NUC']['POS'] > p_snvs[mutations[-1]]['NUC']['POS']:
                reverse_strand = True
            else:
                reverse_strand = False
            # This will become a dict with each mutation being a key and value being an empty list.
            # The list items will correspond to the event seen at the mutational position in each
            # read (0 = No coverage; 1 = REF; 2 = ALT)
            mut_relations = {x: [] for x in mutations}
            # Get the region spanning the group
            if reverse_strand:
                alignment = samfile.fetch(p_snvs[mutations[0]]['NUC']['CHROM'],
                                          p_snvs[mutations[-1]]['NUC']['POS'],
                                          p_snvs[mutations[0]]['NUC']['POS'])
            else:
                alignment = samfile.fetch(p_snvs[mutations[0]]['NUC']['CHROM'],
                                          p_snvs[mutations[0]]['NUC']['POS'],
                                          p_snvs[mutations[-1]]['NUC']['POS'])
            num_spanning_reads = 0
            for read in alignment:
                if read.is_duplicate:
                    continue
                spanned_muts = [x for x in mutations if
                                read.positions[0] <= p_snvs[x]['NUC']['POS'] <= read.positions[-1]]
                for mut in spanned_muts:
                    pos = [x for x, y in read.get_aligned_pairs() if
                           y == p_snvs[mut]['NUC']['POS'] and x is not None]
                    # Make this 0 by default and then modify if we see otherwise. The else clause
                    # for the following 3 if's are handled by this line.
                    mut_relations[mut].append(0)
                    if pos:
                        pos = pos[0]
                        if read.seq[pos] is not None and read.qual[pos] > 30:
                            if p_snvs[mut]['indel'] is None:
                                if read.seq[pos] == p_snvs[mut]['NUC']['ALT']:
                                    mut_relations[mut][-1] = 2
                                elif read.seq[pos] == p_snvs[mut]['NUC']['REF']:
                                    mut_relations[mut][-1] = 1
                            else:
                                # Indel
                                indel_len = (len(p_snvs[mut]['NUC']['ALT']) -
                                             len(p_snvs[mut]['NUC']['REF']))
                                mapped_base_count = 0
                                for ctype, num in read.cigartuples:
                                    if ctype in (0, 4, 7, 8):
                                        # Match or mismatch, or soft clip
                                        mapped_base_count += num
                                        if mapped_base_count > pos + 1:
                                            mut_relations[mut][-1] = 1
                                            break
                                    elif ctype == 1:
                                        # Insertion
                                        if mapped_base_count == pos + 1:
                                            # This could be our event
                                            if num == indel_len:
                                                # This is it
                                                mut_relations[mut][-1] = 2
                                                break
                                            else:
                                                # The only way to handle this location being an
                                                # indel but not the one we are looking for is to
                                                # pretend it doesn't exist.
                                                mut_relations[mut][-1] = 1
                                                break
                                        else:
                                            # We might be looking for an indel further in this read
                                            mapped_base_count += num
                                    elif ctype == 2:
                                        # Deletion
                                        if mapped_base_count == pos + 1:
                                            # This could be our event
                                            if num == -indel_len:
                                                # This is it
                                                mut_relations[mut][-1] = 2
                                                break
                                            else:
                                                # Pretend it doesn't exist.
                                                mut_relations[mut][-1] = 1
                                                break
                                    else:
                                        continue

                for mut in mutations:
                    # Add 0s to the unspanned positions
                    if mut in spanned_muts:
                        continue
                    else:
                        mut_relations[mut].append(0)
                num_spanning_reads += 1
            assert len(mut_relations[mutations[0]]) == num_spanning_reads
            if num_spanning_reads == 0:
                logging.warning('Could not find spanning reads for the mutations (%s). Using all '
                                'combinations', ','.join(mutations))
                out_groups.append(group)
            else:
                # Get a list of how the mutations were phased per read and then process.
                phased_groups = set([tuple([mut_relations[mut][x] for mut in mutations])
                                     for x in xrange(0, num_spanning_reads)])
                # Throw away the useless entries. Useless entries are those that are all 0's (no
                # coverage in the region) and all 1's (No ALT seen in the read)
                useless_groups = [grp for grp in phased_groups
                                  if sum(grp) == 0 or (0 not in grp and 2 not in grp)]
                for grp in useless_groups:
                    phased_groups.remove(grp)
                # Now cleanup the ones with 0's in them.
                # We can fix 0's by phasing with the nearest neighbor from other groups
                final_groups = []
                for grp in phased_groups:
                    if 0 in grp:
                        grp = fix_group(grp, phased_groups)
                        if grp is None:
                            continue
                    final_groups.append([mutations[x] for x, y in enumerate(grp) if y != 1])
                out_groups.append(final_groups)
        logging.debug('Peptide %s detected to have the following mutation groups %s', pfasta_name,
                      out_groups)
        return out_groups


def merge_adjacent_snvs(snvs, muts):
    """
    Given 2-3 mutations that modify the same codon, find the actual modified AA.

    :param dict snvs: The collection of snvs in the protein
    :param list muts: The group of mutations found to be coexpressed
    :return: The corrected tuple of mutations in the group
    :rtype: tuple

    >>> snvs = {'R148R': {'AA': {'change': 'Cgg/Agg'}}, 'R148Q': {'AA': {'change': 'cGg/cAg'}}}
    >>> merge_adjacent_snvs(snvs, ['R148R', 'R148Q'])
    'R148K
    >>> snvs = {'A123B': {'NUC': {'ALT': 'C'}}, \
                'A123D': {'NUC': {'ALT': 'A'}}, \
                'A123F': {'NUC': {'ALT': 'T'}}}
    >>> merge_adjacent_snvs(snvs, ['A123B', 'A123D','A123F'])
    'A123H'
    """
    if len(muts) == 3:
        return muts[0][:-1] + genetic_code[''.join([snvs[x]['NUC']['ALT'] for x in muts])]
    else:
        # Has to be 2 mutations
        changes = zip(snvs[muts[0]]['AA']['change'][4:], snvs[muts[1]]['AA']['change'][4:])
        out_mut = ''
        for change in changes:
            if change[0] == change[1]:
                out_mut += change[0]
            else:
                out_mut += change[0] if change[0].isupper() else change[1]
        return muts[0][:-1] + genetic_code[out_mut.upper()]


def correct_for_same_codon(snvs, mut_group):
    """
    If two mutations in the group modify the same codon, snpeff calls them separately and the naive
    rna filter in Transgene also handles them independently. This method corrects the calls if they
    are found to be on the same allele.

    :param dict(str, dict(str, str|int)) snvs: The collection of snvs in the protein
    :param tuple(stri) mut_group: The group of mutations found to be coexpressed
    :return: The corrected tuple of mutations in the group
    :rtype: tuple(str)

    >>> snvs = {'A123B': {'AA': {'change': 'Xxx/Yxx', 'POS': 123}}, \
                'C124D': {'AA': {'change': 'Zzz/Azz', 'POS': 124}}}
    >>> correct_for_same_codon(snvs, ('A123B',))
    ('A123B',)
    >>> correct_for_same_codon(snvs, ('A123B', 'C124D'))
    ('A123B', 'C124D')
    >>> snvs = {'R148R': {'AA': {'change': 'Cgg/Agg', 'POS': 148}, 'NUC': {'POS': 12345}}, \
                'R148Q': {'AA': {'change': 'cGg/cAg', 'POS': 148}, 'NUC': {'POS': 12346}}, \
                'R149X': {'AA': {'change': 'xXx/xYx', 'POS': 149}, 'NUC': {'POS': 12349}}}
    >>> x = correct_for_same_codon(snvs, ('R148R', 'R148Q', 'R149X'))
    >>> x
    ('R148K', 'R149X')
    >>> 'R148K' in snvs
    True
    >>> snvs['R148K']['AA']['POS'] == 148
    True
    >>> snvs = {'R148R': {'AA': {'change': 'Cgg/Agg', 'POS': 148}, 'NUC': {'POS': 12345}}, \
                'R148W': {'AA': {'change': 'Cgg/Tgg', 'POS': 148}, 'NUC': {'POS': 12345}}, \
                'R149X': {'AA': {'change': 'xXx/xYx', 'POS': 149}, 'NUC': {'POS': 12349}}}
    >>> correct_for_same_codon(snvs, ('R148R', 'R148W', 'R149X'))
    ()
    """
    if len(mut_group) == 1 or len(mut_group) == len({snvs[x]['AA']['POS'] for x in mut_group}):
        # If there is only 1 mutation or if the positions are unique
        return mut_group
    else:
        out_list = []
        # A  set of positions in this group
        positions = {snvs[x]['AA']['POS'] for x in mut_group}
        # These groups are going to be small so we aren't losing much by brute-forcing this.
        for pos in sorted(positions):
            muts = [mut for mut in mut_group if str(pos) in mut]
            if len(muts) == 1:
                out_list.extend(muts)
            elif len({snvs[mut]['NUC']['POS'] for mut in muts}) == 1:
                # MAVs affect the same codon AND the same nucleotide. They CANNOT be in the same
                # group
                logging.info('Skipping a group (%s) containing a Multi-Allelic Variant.', muts)
                return ()
            else:
                logging.info('Identified co-expressed mutations (%s)that affect the same '
                             'codon/AA.  Merging.', muts)
                merged_group = merge_adjacent_snvs(snvs, muts)
                out_list.append(merged_group)
                logging.info('Merged (%s) into %s.', muts, merged_group)
                # Now add this to the mutations dict
                snvs[merged_group] = {'AA': {'REF': merged_group[0],
                                             'ALT': merged_group[-1],
                                             'POS': pos,
                                             'VAF': min([snvs[x]['AA']['VAF'] for x in mut_group])},
                                      'indel': False}
        return tuple(out_list)


def insert_mutations(protein_fa, mutations, tumfile, normfile, peplen, rna_bam=None, chroms=None,
                     exons=None, cds_starts=None, extend_length=10):
    """
    This module uses the mutation data contained in `mutations` and inserts them into the
    genome contained in `chroms`.

    :param dict protein_fa: Contains the peptides in the form of a dict where keys hold the protein
           name and the values holds the sequence.
    :param dict mutations: Contains the snvs parsed from teh input SnpEFF file.
    :param file tumfile: The file to write tumor output to.
    :param file normfile: The file to write normal output to.
    :param int peplen: Length of peptides which will be generated from the output file.
    :param str rna_bam: The path to the RNA-Seq bam file
    :param dict chroms: Contains the chromosomal dna in the form of a dict where keys hold the
           chromosome name and the values holds the sequence.
    :param dict exons: See return value of `get_exons`
    :param dict cds_starts: See return value of `get_exons`
    :param int extend_length: The number of codons downstream of an indel to process
    """
    logging.info('Inserting Mutations into IARs')
    for pept in mutations.keys():
        # First, grab the positions of all mutations in the peptide.  If there are multiple
        # mutations, then for each mutation, list the other mutations that are within peplen
        # positions of it IN THE POSITIVE DIRECTION.  The mutations are potentially co-expressed on
        # the same peptide. The reason can be visually shown for a 9-mer peptide with three
        # mutations, X, Y and Z as below (O = non-mutated AA).  The region depicts the extended IAR
        # for the 3. Possible mutational combinations are shown. The WT (O,O,O) isn't shown since it
        # isn't useful.
        #
        # POSITIONS:    a b c d e f g h i j k l m n o p q r s t u v w x y
        #
        # X+Y+Z        O O O O O O O O X O O O Y O O O Z O O O O O O O O
        # X + Z        O O O O O O O O X O O O O O O O Z O O O O O O O O
        # X + Y        O O O O O O O O X O O O Y O O O O O O O O O O O O
        # X only       O O O O O O O O X O O O O O O O O O O O O O O O O
        #
        # Y + Z        O O O O O O O O O O O O Y O O O Z O O O O O O O O
        # Y only       O O O O O O O O O O O O Y O O O O O O O O O O O O
        #
        # Z only       O O O O O O O O O O O O O O O O Z O O O O O O O O
        #
        # The data for what mutations are grouped together, and in what combinations is stored in
        # the mutation_groups list object. Each group in in the list is a list of all mutations in
        # the group.  If there are no neighboring mutations, the group contains only the mutation
        # itself.
        pfasta_name = [x for x in protein_fa.keys() if ''.join([pept, '_']) in x][0]
        # We have all the mutations and the protein sequence here. Do a sanity check on the results.
        protein = protein_fa[pfasta_name]
        if 'len' in mutations[pept]:
            # If 'len' is not in this dict, it means a previous process has already cleaned up the
            # dict. This will need to change if this module is multithreaded.
            expected_tlen = mutations[pept].pop('len')
            mutations[pept].pop('HUGO')  # This has served its purpose
            for mutation in mutations[pept].keys():
                if mutation == 'has_indel':
                    # Non mutation key
                    continue
                if mutations[pept][mutation]['indel']:
                    # There is no need to correct indels
                    continue
                mut_pos = mutations[pept][mutation]['AA']['POS']
                if protein[mut_pos - 1] != mutations[pept][mutation]['AA']['REF']:
                    if mutations[pept][mutation]['AA']['REF'] == \
                            mutations[pept][mutation]['AA']['ALT']:
                        # Let synonymous mutations pass through
                        continue
                    logging.warning('%s seen at position %s in %s.... %s expected.',
                                    protein_fa[pfasta_name][mut_pos - 1], mut_pos, pept,
                                    mutations[pept][mutation]['AA']['REF'])
                    if len(protein_fa[pfasta_name]) - expected_tlen in (1, 2):
                        # This is a case seen often with SNPEff where the n+1 position will have
                        # the mutation.  Check the n+1 residue
                        if protein[mut_pos] == mutations[pept][mutation]['AA']['REF']:
                            logging.info('Mutation at position %s in %s detected to exist at '
                                         'position %s.', mut_pos, pept, mut_pos + 1)
                            new_mutation = ''.join([mutations[pept][mutation]['AA']['REF'],
                                                    str(mut_pos + 1),
                                                    mutations[pept][mutation]['AA']['ALT']])
                            mutations[pept][new_mutation] = mutations[pept][mutation]
                            mutations[pept][new_mutation]['AA']['POS'] = mut_pos + 1
                            mutations[pept].pop(mutation)
                    else:
                        logging.warning('Cannot handle mutation at position %s in %s. '
                                        'Disregarding.', mut_pos, pept)
                        mutations[pept].pop(mutation)
        if not mutations[pept]:
            continue

        all_mutation_groups = get_mutation_groups(mutations[pept], peplen, pfasta_name, rna_bam)

        for mut_group in all_mutation_groups:
            peptide_muts = deepcopy(mutations)
            for group_muts in mut_group:
                group_muts = correct_for_same_codon(peptide_muts[pept], group_muts)
                # After correcting, if there is only ONE mutation and it is a stop gain, disregard
                # this call
                if len(group_muts) == 1 and group_muts[0][-1] == '*':
                    continue
                group_muts = tuple([x for x in group_muts
                                    if peptide_muts[pept][x]['AA']['REF'] !=
                                    peptide_muts[pept][x]['AA']['ALT']])
                if len(group_muts) == 0:
                    # If the group only consisted of one synonymous mutation, or was rejected for
                    # having a MAV.
                    continue
                out_pept = {'pept_name': [pfasta_name],
                            'tum_seq': [],
                            'norm_seq': []
                            }
                mut_pos = peptide_muts[pept][group_muts[0]]['AA']['POS']
                prev = max(mut_pos - peplen, 0)  # First possible AA in the IAR
                skip = False  # A flag to identify we need to skip this group
                for group_mut in group_muts:
                    mut_pos = peptide_muts[pept][group_mut]['AA']['POS']
                    ref = peptide_muts[pept][group_mut]['AA']['REF']
                    alt = peptide_muts[pept][group_mut]['AA']['ALT']
                    out_pept['pept_name'].append(group_mut + '#' +
                                                 peptide_muts[pept][group_mut]['AA']['VAF'])
                    out_pept['tum_seq'].extend(protein[prev:mut_pos - 1])
                    out_pept['norm_seq'].extend(protein[prev:mut_pos - 1])
                    if '*' in alt:
                        star_pos = alt.find('*')
                        # Append any AAs before a stop codon in an frame-shift induced neoepitope
                        out_pept['tum_seq'].append(alt[:star_pos])
                        prev = None
                        break
                    elif not peptide_muts[pept][group_mut]['indel']:
                        # SNV. Easy to handle
                        out_pept['tum_seq'].append(alt)
                        out_pept['norm_seq'].append(ref)
                        prev = mut_pos
                    else:
                        # Indels
                        if 'frame_shift' not in peptide_muts[pept][group_mut]['indel']:
                            # Full codon insertion or insertion with codon change, or full codon
                            # deletion.  We assume none of these events affect splicing.
                            if ref == '-':
                                # VEP represents an in-frame insertion as -/Q where a Q is inserted
                                # with no change to the surrounding sequence
                                out_pept['tum_seq'].append(protein[mut_pos - 1])
                            out_pept['tum_seq'].extend(alt)
                            out_pept['norm_seq'].extend(protein[mut_pos - 1:mut_pos - 1 + len(ref)])
                            prev = mut_pos + len(ref) - 1
                        else:
                            # Frame shift mutation.
                            group_mut_pos = group_muts.index(group_mut)
                            aa_seq = get_genomic_seq(pept, group_muts[group_mut_pos:],
                                                     peptide_muts[pept], chroms, exons, cds_starts,
                                                     extend_length)
                            if aa_seq is None:
                                # aa_seq will be a string of characters for a frameshift mutant
                                # and None if it was a frameshift mutant in a gene that
                                # has no known CDS start.  We can't handle the second case.
                                skip = True
                                logging.warning('(%s) has no known CDS start. Skipping.',
                                                out_pept['pept_name'])
                                break
                            out_pept['tum_seq'].extend(aa_seq)
                            # Append the immediate next normal amino acids in the chain as long as
                            # they match the translated sequence. We add one mismatch and then
                            # terminate the normal strand. We do this in case we need to trim the
                            # sequences.
                            for pos, aa in enumerate(aa_seq):
                                if pos + mut_pos > len(protein):
                                    # Reached the end of the protein
                                    break
                                out_pept['norm_seq'].append(protein[mut_pos + pos - 1])
                                if aa != protein[mut_pos + pos - 1]:
                                    break
                            out_pept['norm_seq'].append('#')  # Terminate normal strand
                            prev = None
                        break

                if skip:
                    break
                if prev is not None:
                    out_pept['tum_seq'].extend(protein[prev:prev + peplen - 1])
                    out_pept['norm_seq'].extend(protein[prev:prev + peplen - 1])
                if peptide_muts[pept]['has_indel']:
                    # Handle left side of the IAR
                    fm = first_mismatch(out_pept['tum_seq'], out_pept['norm_seq'])
                    if fm >= peplen:
                        for _ in range(peplen - 1, fm):
                            out_pept['tum_seq'].pop(0)
                            out_pept['norm_seq'].pop(0)
                    elif fm == -1:
                        # This will happen if they are the same sequence after translating. Discard
                        # This sequence
                        logging.warning('IAR in (%s) is exactly the same as its normal counterpart '
                                        '(T: %s, N: %s). Skipping.', out_pept['pept_name'],
                                        ''.join(out_pept['tum_seq']), ''.join(out_pept['norm_seq']))
                        continue
                    # Handle right side of the IAR
                    fm = first_mismatch(out_pept['tum_seq'][::-1], out_pept['norm_seq'][::-1])
                    if fm >= peplen:
                        for _ in range(peplen - 1, fm):
                            out_pept['tum_seq'].pop()
                            out_pept['norm_seq'].pop()
                write_pepts_to_file(out_pept, tumfile, normfile, peplen)
    return None


def get_genomic_seq(pept, group_muts, pept_muts, chroms, exons, cds_starts, extend_length):
    """
    Get the genomic sequence downstream to a mutation containing all other chained mutations.

    :param str pept: The peptide name
    :param tuple group_muts: Group mutations
    :param dict pept_muts: All mutations for the peptide
    :param dict chroms: see insert_mutations:`chroms`
    :param dict exons: see insert_mutations:`exons`
    :param dict cds_starts: see insert_mutations:`cds_starts`
    :param int extend_length: see insert_mutations:`extend_length`
    :return:
    """
    chrom = pept_muts[group_muts[0]]['NUC']['CHROM']
    first_mut = group_muts[0]
    # curr_codon_start_pos will give the start position of the codon immediately BEFORE the mutation
    curr_codon_start_pos, positive_strand = get_exon_start_pos(
        pept, pept_muts[first_mut]['AA']['POS'], exons, cds_starts)
    if curr_codon_start_pos is None:
        # This gene doesn't have a known CDS so we can't handle it.
        return None
    curr_codon_start_pos -= 1  # The value obtained from the gtf is 1-based

    codon_extend_length = extend_length * 3  # +1 for the current codon

    if positive_strand:
        # If the indel started at a new codon, the first base in the insertion or the retained base
        # in the deletion actually belongs to the previous codon reported by annotators since the
        # first "Affected" codon is reported.  This is a problem only on the positive strand
        if curr_codon_start_pos == pept_muts[first_mut]['NUC']['POS'] + 1:
            if (len(pept_muts[first_mut]['NUC']['ALT']) >
                    len(pept_muts[first_mut]['NUC']['REF'])):
                # In an insertion, since we just dump the contents of alt into the tumor seq, we can
                # just modify alt to remove that first base
                pept_muts[first_mut]['NUC']['ALT'] = pept_muts[first_mut]['NUC']['ALT'][1:]
            else:
                # In a deletion, we appned the alt allele and extend by len(ref) - len(alt). We can
                # fix this by trimming both ref and alt.
                pept_muts[first_mut]['NUC']['ALT'] = ''
                pept_muts[first_mut]['NUC']['REF'] = pept_muts[first_mut]['NUC']['REF'][1:]

    next_n_bases = []
    prev = curr_codon_start_pos
    for mut in group_muts:
        mut_pos = pept_muts[mut]['NUC']['POS']
        if abs(mut_pos - curr_codon_start_pos) > extend_length:
            break
        if positive_strand:
            next_n_bases.extend(chroms[chrom][prev:mut_pos])
        else:
            next_n_bases.extend(chroms[chrom][mut_pos + 1:prev + 1][::-1])
        if len(pept_muts[mut]['NUC']['REF']) <= len(pept_muts[mut]['NUC']['ALT']):
            # Not a deletion. Just insert the alt allele
            if positive_strand:
                next_n_bases.append(pept_muts[mut]['NUC']['ALT'])
                prev = mut_pos + 1
            else:
                next_n_bases.append(pept_muts[mut]['NUC']['ALT'][::-1])
                prev = mut_pos - 1
            codon_extend_length -= (len(pept_muts[mut]['NUC']['ALT']) - 1)
        else:
            # Deletion
            del_length = len(pept_muts[mut]['NUC']['REF']) - len(pept_muts[mut]['NUC']['ALT'])
            if positive_strand:
                prev = mut_pos + del_length + 1
            else:
                for _ in range(del_length):
                    next_n_bases.pop()
                prev = mut_pos - 1
            codon_extend_length += (len(pept_muts[mut]['NUC']['REF']) - 1)
            next_n_bases.append(pept_muts[mut]['NUC']['ALT'])
    if positive_strand:
        next_n_bases.extend(chroms[chrom][prev:curr_codon_start_pos + codon_extend_length])
        next_n_bases = ''.join(next_n_bases)
    else:
        next_n_bases.extend(chroms[chrom][curr_codon_start_pos -
                                          codon_extend_length:prev + 1][::-1])
        next_n_bases = ''.join(next_n_bases).translate(trans)
    genomic_seq = translate(next_n_bases)
    if '*' in genomic_seq:
        genomic_seq = genomic_seq[:genomic_seq.find('*')]

    return genomic_seq


def write_pepts_to_file(peptide_info, tumfile, normfile, peplen):
    """
    This module will accept the info for a given peptide including the name, the tumor peptide
    sequence, and the corresponding normal peptide sequence, and will print then to tumfile and
    normfile
    """
    blacklist = set(list('BJOUXZ'))
    peptide_name = '_'.join(peptide_info['pept_name'])
    tum_peptide_seq = ''.join(peptide_info['tum_seq'])
    norm_peptide_seq = ''.join(peptide_info['norm_seq'])
    if '#' in norm_peptide_seq:
        norm_peptide_seq = norm_peptide_seq[:norm_peptide_seq.find('#')]
        if len(norm_peptide_seq) < peplen:
            logging.warning('Corresponding normal IAR for %s is less than %s residues'
                            '(%s).', peptide_name, peplen, norm_peptide_seq)
            norm_peptide_seq = ''
    truncated = False
    if not blacklist.isdisjoint(tum_peptide_seq):
        # If the final peptide contains a blacklisted amino acid,
        # handle it.
        logging.warning('Blacklisted amino acid(s) %s seen in IAR %s... Truncating.',
                        list(set(list(tum_peptide_seq)).intersection(blacklist)), peptide_name)
        matches = [i for i, x in enumerate(tum_peptide_seq) if x in blacklist]
        truncated = True
        if len(matches) > 1:
            # Out of 94285 transcripts in gencode 25, Only 14 have multiple blacklisted AAs. Of
            # these 14, only 6 transcripts from 4 genes (3 + 1 + 1 + 1) have the possibility of >2
            # blacklisted AAs being in the same single-mutation 9-, 10- and 15-mer, and double 9-
            # and 10-mers. Another transcript from another gene (total 7) has a possibility of >2
            # blacklisted AAs in a  double-mutation 15-mer.  It seems logically ok to just disregard
            # this edge case for now.
            logging.warning('Cannot handle multiple blacklisted AAs per IAR.')
            return None
        else:
            # If there is one truncation point, you get 2 piece of either the same size (necessarily
            # both are less than peplen in size), or dissimilar size where if the larger piece is
            # greater than or equal to peplen in size, it must necessarily contain the mutant.
            # Pieces smaller than peplen are neglected.
            pos = matches[0]
            if pos > len(tum_peptide_seq)/2:
                tum_peptide_seq = tum_peptide_seq[:pos]
                norm_peptide_seq = norm_peptide_seq[:pos]
            else:
                tum_peptide_seq = tum_peptide_seq[pos+1:]
                norm_peptide_seq = norm_peptide_seq[pos+1:]

    if len(tum_peptide_seq) < peplen:
        prefix = 'Truncated ' if truncated else ''
        logging.warning((prefix + 'IAR %s is less than %s residues (%s).'), peptide_name, peplen,
                        tum_peptide_seq)
    else:
        print('>', peptide_name, sep='', file=tumfile)
        print(tum_peptide_seq, file=tumfile)
        if norm_peptide_seq:
            print('>', peptide_name, sep='', file=normfile)
            print(norm_peptide_seq, file=normfile)
    return None


def parse_peptides(tumfile, normfile, prefix, suffix):
    """
    This module takes in a peptides file and squashes it into the minimum number of peptides
    required to describe the potential neo-immunopeptidome.  It's main function is to take
    transcript-level mutation calls and merge them by gene if the sequences they contain are
    identical (to correct for splicing information).
    """
    with open(normfile, 'r') as i_f:
        normal_peptides = {}
        for pep_name, _, pep_seq in read_fasta(i_f, 'ARNDCQEGHILKMFPSTWYVBZJUOX'):
            normal_peptides[pep_name] = pep_seq
    with open(tumfile, 'r') as i_f:
        peptides = collections.Counter()
        for full_pep_name, _, pep_seq in read_fasta(i_f, 'ARNDCQEGHILKMFPSTWYVBZJUOX'):
            pep_name = full_pep_name.split('_')
            gene_name = pep_name[0]
            hugo_gene = pep_name[2]
            transcript_mutation = '_'.join([pep_name[1]]+pep_name[3:])
            if peptides[(gene_name, hugo_gene)] == 0:
                try:
                    peptides[(gene_name, hugo_gene)] = {transcript_mutation: (
                        pep_seq, normal_peptides[full_pep_name])}
                except KeyError:
                    peptides[(gene_name, hugo_gene)] = {transcript_mutation: (pep_seq, None)}
            else:
                try:
                    peptides[(gene_name, hugo_gene)][transcript_mutation] = (
                        pep_seq, normal_peptides[full_pep_name])
                except KeyError:
                    peptides[(gene_name, hugo_gene)][transcript_mutation] = (pep_seq, None)
    outmap = {}
    out_tum_file = '_'.join([prefix, 'tumor', suffix])
    out_norm_file = '_'.join([prefix, 'normal', suffix])
    with open(out_tum_file, 'w') as t_f, open(out_norm_file, 'w') as n_f:
        # Mhc predictors can't handle the giant peptide names created by merging the names of
        # peptide emerging from transcripts bearing similar mutations hence we use an incremental
        # system to name the peptides and store the peptide number to actual peptide name
        # information in a map file.
        peptide_number = 1
        for gene in peptides.keys():
            unique_seqs = set(peptides[gene].values())
            for group in [[x for x, y in peptides[gene].items() if y[0] == z[0]]
                          for z in unique_seqs]:
                pepname = ''.join(['neoepitope_', str(peptide_number)])
                group_info = '\t'.join(list(gene) + [','.join(group)])
                group_tum_seq = peptides[gene][group[0]][0]
                group_norm_seq = peptides[gene][group[0]][1]
                # Print to faa files
                print('>', pepname, sep='', file=t_f)
                print(group_tum_seq, file=t_f)

                if group_norm_seq is not None:
                    print('>', pepname, sep='', file=n_f)
                    print(group_norm_seq, file=n_f)
                # Save to map dict
                outmap[pepname] = group_info
                peptide_number += 1
    #  Json dump the results to file
    with open(''.join([out_tum_file, '.map']), 'w') as mapfile:
        json.dump(outmap, mapfile)
    return None


def get_proteome_data(infile, drop_transcript_version=False):
    """
    Loads GENCODE translated transcript sequences into a dictionary

    :param file infile: GENCODE file object
    :return: Mapping of concatenated Ensembl gene IDs, Ensembl transcript IDs, and HUGO names to
             protein sequences
    :param bool drop_transcript_version: Drop the version part of transcript ids.
    :rtype: dict
    """
    # Read the proteomic fasta
    logging.info('Reading GENCODE proteome fasta file')
    peptides = collections.Counter()
    for fa_seq in read_fasta(infile, 'ARNDCQEGHILKMFPSTWYVBZJUOX'):
        #  Fastq headers are ALWAYS 7 or 8 |-separated fields long. The fields are
        #  0a. Ensembl Peptide (optional) -- e.g ENSP00000334393.3
        #  0b. Ensembl Transcript         -- e.g ENST00000335137.3
        #  1. Ensembl Gene                -- e.g ENSG00000186092.4
        #  2. Havana Gene                 -- e.g OTTHUMG00000001094.2
        #  3. Havana Transcript           -- e.g OTTHUMT00000003223.2
        #  4. HGNC gene splice variant    -- e.g OR4F5-001
        #  5. HUGO name / HGNC symbol     -- e.g OR4F5
        #  6. Length in AA residues       -- e.g 305
        #  We need columns 0b, 1,  and 5
        try:
            record_name = list(itemgetter(-6, -7, -2)(fa_seq[0].split('|')))
            if drop_transcript_version:
                record_name[1] = record_name[1].split('.')[0]
            record_name = '_'.join(record_name)
        except IndexError:
            raise RuntimeError('Was the input peptides file obtained from Gencode?')
        peptides[record_name] = list(fa_seq[2])
    return peptides


def main(params):
    """
    Transgene accepts a SnpEffed vcf of translated mutations -- i.e. a file that has information of
    the mutation in genomic space (e.g. chr1:47403668) and in proteomic space (e.g.
    ENST00000371904.4:D113N) -- and outputs files that contain n-mer peptides (n>0) that can be
    passed to MHC:peptide binding prediction tools. Additionally, this tool accepts gene fusion
    annotations in BEDPE format and outputs n-mer peptides across the gene fusion boundary.

    Transgene can be run in an RNA-aware mode where it will only accept small mutations that have
    evidence in an RNA-seq bam (Mapped to the genome reference). It does not look for evidence of
    fusion events in the RNA-seq bam.
    """
    start_time = time.time()
    if params.logfile is not None:
        params.logfile = os.path.abspath(os.path.expanduser(params.logfile))
        assert os.path.exists(os.path.dirname(params.logfile)), 'Invalid Directory to write logfile'
        print('Logging to file: %s' % params.logfile)
    logging.basicConfig(filename=params.logfile,
                        level=getattr(logging, params.log_level),
                        format='%(levelname)s: %(message)s')

    # Process the arguments
    if params.rna_file is not None:
        params.rna_file = os.path.abspath(os.path.expanduser(params.rna_file))
        if not os.path.exists(params.rna_file):
            raise RuntimeError('Could not file the RNA-seq bam file at the provided location.')
        if not (os.path.exists(params.rna_file + '.bai') or
                    os.path.exists(params.rna_file[:-1] + 'i')):
            # /a/b/c/def.bam -> /a/b/c/def.bam.bai or /a/b/c/def.bai
            raise RuntimeError('Could not file the RNA-seq bai file at the provided location.')

    if params.dna_file is not None:
        params.dna_file = os.path.abspath(os.path.expanduser(params.dna_file))
        if not os.path.exists(params.dna_file):
            raise RuntimeError('Could not file the input tumor DNA-seq bam file at the provided '
                               'location.')
        if not (os.path.exists(params.dna_file + '.bai') or
                    os.path.exists(params.dna_file[:-1] + 'i')):
            raise RuntimeError('Could not file the input tumor DNA-seq bai file at the provided '
                               'location.')

    # Load proteomic data
    proteome_data = None
    if params.peptide_file:
        proteome_data = get_proteome_data(params.peptide_file, params.vep_file is not None)
        params.peptide_file.close()

    # Read in annotated vcf files
    mutations = None
    genes_to_translate = set()
    annotator = None
    if params.snpeff_file or params.vep_file:
        if params.rna_file or params.dna_file:
            out_vcf = open('_'.join([params.prefix, 'transgened.vcf']), 'w')
        else:
            out_vcf = None
        try:
            mutations, annotator = read_annotated_vcf(vcf_file=(params.snpeff_file or
                                                                params.vep_file),
                                                      rna_file=params.rna_file,
                                                      out_vcf=out_vcf,
                                                      reject_threshold=params.reject_threshold,
                                                      rna_min_alt_freq=params.rna_min_alt_freq,
                                                      dna_file=params.dna_file,
                                                      oxog_min_alt_freq=params.oxog_min_alt_freq,
                                                      processes=params.cores,
                                                      reject_indels=not(params.genome_file and
                                                                        params.annotation_file))
        except:
            raise
        else:
            if annotator not in ('SNPEff3.6', 'SNPEff4.0', 'VEP'):
                # SNPEff_4.1+ does not give us the exact codon changes (Aag/Gat) instead it just
                # gives us the changed Nucleotide in cDNA space (p.A>G).  We need to translate all
                # genes now so we can get the actual codon changes.
                genes_to_translate.update([mutations[m]['HUGO'] for m in mutations])
                temp_string = annotator + ' annotated vcfs'
            else:
                genes_to_translate.update([mutations[m]['HUGO']
                                           for m in mutations if mutations[m]['has_indel']])
                temp_string = 'indels'
            if genes_to_translate and not (params.genome_file and params.annotation_file):
                raise RuntimeError('Cannot handle %s without a genome fasta and a gencode '
                                   'annotation.' % temp_string)
        finally:
            if params.rna_file or params.dna_file:
                out_vcf.close()
            # These files were opened by argparse
            if params.snpeff_file:
                params.snpeff_file.close()
            else:
                params.vep_file.close()

    # Load data for generating fusion peptides
    if params.transcript_file:
        transcriptome, gene_transcript_ids = get_transcriptome_data(params.transcript_file,
                                                                    params.vep_file is not None)
    else:
        transcriptome, gene_transcript_ids = None, None

    # Load data from fusion file
    fusions = exons = cds_starts = None
    if params.fusion_file:
        gene_annotations = read_genes_from_gtf(params.annotation_file)
        out_bedpe = open('_'.join([params.prefix, 'transgened.bedpe']), 'w')
        try:
            fusions = read_fusions(params.fusion_file, gene_annotations, params.filter_mt,
                                   params.filter_ig, params.filter_rg, params.filter_rt,
                                   params.rt_threshold, out_bedpe)
        finally:
            out_bedpe.close()
        genes_to_translate.update(sum([(record.hugo1, record.hugo2) for record in fusions], ()))

    if not (fusions or mutations):
        logging.warning('The input files had no actionable mutations.')
        logging.info('Transgene successfully ran in %s seconds', round(time.time() - start_time, 2))
        exit(0)
    chroms = {}
    if genes_to_translate:
        exons, cds_starts = get_exons(params.genome_file, params.annotation_file,
                                      genes_to_translate,
                                      params.vep_file is not None)

        chroms_of_interest = {e[0].seqname for e in exons.values()}
        params.genome_file.seek(0)
        for header, comment, seq in read_fasta(params.genome_file, 'ACGTN'):
            if header in chroms_of_interest:
                chroms[header] = seq

    # SNPEff 4.1+ does not give use the codon change so before we can start inserting mutations, we
    # need to fill in the blanks.
    if annotator and annotator not in ('SNPEff3.6', 'SNPEff4.0', 'VEP'):
        not_in_annot = set()
        for transcript in mutations:
            for mutation in mutations[transcript]:
                if mutation in ('HUGO', 'has_indel', 'len'):
                    continue
                else:
                    if mutations[transcript][mutation]['indel'] is None:
                        temp_mut = mutations[transcript][mutation]
                        codon = get_codon(transcript,  temp_mut['AA']['POS'], exons, cds_starts,
                                          chroms[mutations[transcript][mutation]['NUC']['CHROM']])
                        if codon is None:
                            not_in_annot.add(transcript)
                            logging.warning('Transcript %s was not found in the given GTF, or did '
                                            'not have a registered CDS start. Skipping.',
                                            transcript)
                            break
                        mutated_pos = int(temp_mut['AA']['change'][2:-3])
                        mutated_pos = (mutated_pos - 1) % 3
                        codon_change = temp_mut['AA']['change'][-3], temp_mut['AA']['change'][-1]
                        assert codon[mutated_pos] == codon_change[0]
                        outstring = list((codon + '/' + codon).lower())
                        outstring[mutated_pos], outstring[4 + mutated_pos] = codon_change
                        mutations[transcript][mutation]['AA']['change'] = ''.join(outstring)
        for transcript in not_in_annot:
            mutations.pop(transcript)

    for peplen in params.pep_lens.split(','):
        logging.info('Processing %s-mers', peplen)
        outfile1, tumfile_path = mkstemp()
        outfile2, normfile_path = mkstemp()
        os.close(outfile1)
        os.close(outfile2)

        with open(tumfile_path, 'w') as tumfile, open(normfile_path, 'w') as normfile:
            if proteome_data and mutations:
                insert_mutations(proteome_data, mutations, tumfile, normfile, int(peplen),
                                 params.rna_file, chroms=chroms, exons=exons, cds_starts=cds_starts,
                                 extend_length=params.extend_length)
            if transcriptome and gene_transcript_ids and fusions:
                insert_fusions(transcriptome, fusions, gene_transcript_ids, int(peplen), tumfile,
                               exons=exons)

        if params.no_json_dumps:
            shutil.move(tumfile_path, '_'.join([params.prefix, 'tumor', peplen,
                                                'mer_peptides.faa']))
            shutil.move(normfile_path, '_'.join([params.prefix, 'normal', peplen,
                                                 'mer_peptides.faa']))
        else:
            parse_peptides(tumfile_path, normfile_path, params.prefix,
                           '_'.join([peplen, 'mer_peptides.faa']))
            os.remove(tumfile_path)
            os.remove(normfile_path)

    # Temporary file used during fusion alignment steps
    if os.path.exists('transgene_fusion_alignments.pkl'):
        os.remove('transgene_fusion_alignments.pkl')

    logging.info('Transgene successfully ran in %s seconds', round(time.time() - start_time, 2))


def run_transgene():
    """
    This will try to run transgene from system arguments
    """
    parser = argparse.ArgumentParser(description=main.__doc__)
    # Mutation related options
    parser.add_argument('--peptides', dest='peptide_file', type=argparse.FileType('r'),
                        help='Path to GENCODE translation FASTA file')
    vcf_file = parser.add_mutually_exclusive_group()
    vcf_file.add_argument('--snpeff', dest='snpeff_file', type=argparse.FileType('r'),
                          help='Path to snpeff file', default=None)
    vcf_file.add_argument('--vep', dest='vep_file', type=argparse.FileType('r'),
                          help='Path to vep file', default=None)
    parser.add_argument('--genome', dest='genome_file',
                        help='Path to reference genome fasta file, required if calling fusions, '
                        'indels, or SNPEff 4.1+.', type=file_type, required=False, default=None)
    parser.add_argument('--indel_extend_length', dest='extend_length',
                        help='The number of Codons downstream to an indel to process.',
                        type=int, default=10)
    # Fusion related options
    parser.add_argument('--fusions', dest='fusion_file', help='Path to gene fusion file',
                        type=argparse.FileType('r'))
    parser.add_argument('--transcripts', dest='transcript_file', type=argparse.FileType('r'),
                        help='Path to GENCODE transcript FASTA file. Required if calling fusions.')
    parser.add_argument('--annotation', dest='annotation_file',
                        help='Path to gencode annotation file. Required if calling fusions or if'
                        'using SNPEff 4.1+.',
                        type=file_type)
    parser.add_argument('--filter_mt_fusions', dest='filter_mt', action='store_true',
                        help='Filter fusions involving Mitochondrial genes.', required=False)
    parser.add_argument('--filter_ig_pairs', dest='filter_ig', action='store_true',
                        help='Filter fusions involving two immunoglobulin genes (IGXXX).',
                        required=False)
    parser.add_argument('--filter_rna_gene_fusions', dest='filter_rg', action='store_true',
                        help='Filter fusions involving RNA genes (RP11-XXXX).', required=False)
    parser.add_argument('--filter_readthroughs', dest='filter_rt', action='store_true',
                        help='Filter transcriptional read-troughs.', required=False)
    parser.add_argument('--readthrough_threshold', dest='rt_threshold', type=int,
                        help='Genomic distance between candidates on the same strand below which a '
                        'fusion will be considered a read-through.', default=500000, required=False)

    # RNA-Aware options
    parser.add_argument('--rna_file', dest='rna_file', help='The path to an RNA-seq bam file. If '
                        'provided, the vcf will be filtered for coding mutations only. The file '
                        'must be indexed with samtools index.',
                        required=False, default=None)
    parser.add_argument('--reject_threshold', dest='reject_threshold', help='The minimum number of '
                        'reads containing the REF allele required to reject an event where no '
                        'reads contain the ALT allele.  If the value is lower than this, we will '
                        'accept the mutation.', type=int, required=False, default=5)
    parser.add_argument('--min_rna_alt_freq', dest='rna_min_alt_freq', help='The ALT allele '
                        'frequency (as a fraction) in the RNA-Seq below which we will reject the '
                        'mutation.', type=float, required=False, default=0.1)

    # OxoG filtering options
    parser.add_argument('--filterOxoG', dest='filter_oxog', action='store_true', help='Filter the '
                        'calls for OxoG artifacts. This feature requires a tumor dna bam as input.',
                        required=False, default=False)
    parser.add_argument('--dna_file', dest='dna_file', help='The path to an tumor DNA-seq bam '
                        'file. This is required for OxoG artifact filtering. The file must be '
                        'indexed with samtools index.', required=False, default=None)
    parser.add_argument('--min_OxoG_variant_freq', dest='oxog_min_alt_freq', help='The ALT '
                        'allele frequency (as a fraction) in the DNA-Seq below which we will flag'
                        'the mutation as being an OxoG variant.',
                        type=float, required=False, default=0.1)

    # Logging
    parser.add_argument('--log_level', dest='log_level', help='The level of logging above which '
                        'messages should be printed.', required=False, choices={'DEBUG', 'INFO',
                                                                                'WARNING', 'ERROR'},
                        default='INFO')
    parser.add_argument('--log_file', dest='logfile', help='A path to a logfile.', type=str,
                        required=False, default=None)

    # Misc
    parser.add_argument('--prefix', dest='prefix', type=str, help='Prefix for output file names.',
                        required=True)
    parser.add_argument('--pep_lens', dest='pep_lens', type=str, help='Desired peptide lengths to '
                        'process. The argument should be in the form of comma separated values.  '
                        'E.g. 9,15', required=False, default='9,10,15')
    parser.add_argument('--no_json_dumps', action='store_true',
                        help='Do not educe peptide fasta record names in the output by dumping the '
                        'mapping info into a .map json file.', required=False, default=False)
    parser.add_argument('--cores', dest='cores', type=int,
                        help='Number of cores to use for the filtering step.', required=False,
                        default=1)
    params = parser.parse_args()

    if (params.snpeff_file or params.vep_file) and not params.peptide_file:
        raise ValueError('VCF file requires GENCODE translation FASTA file')

    if params.fusion_file and not params.transcript_file:
        raise ValueError('Fusion calling requires GENCODE transcripts FASTA file')
    if params.fusion_file and not params.annotation_file:
        raise ValueError('Fusion calling requires GENCODE gtf annotation file')
    if params.fusion_file and not params.genome_file:
        raise ValueError('Fusion calling requires genomic fasta')

    if params.filter_oxog:
        if not params.dna_file:
            raise ValueError('OxoG artifact filtering required a tumor DNA-Seq bam file.')
        if params.oxog_min_alt_freq > 1.0 or params.oxog_min_alt_freq < 0.0:
            raise ValueError('--min_OxoG_variant_threshold must be a fraction between 0 and 1.')

    if params.rna_file:
        if params.rna_min_alt_freq > 1.0 or params.rna_min_alt_freq < 0.0:
            raise ValueError('--min_rna_alt_threshold must be a fraction between 0 and 1.')

    return main(params)


if __name__ == '__main__':
    sys.exit(run_transgene())



import collections
import gzip
import os
import re


# A translation table use to complement string sequences
import string

forward = 'ACGTN'
reverse = 'TGCAN'
trans = string.maketrans(forward, reverse)


def read_fasta(input_file, alphabet):
    """
    This module will accept an input fasta and yield individual sequences
    one at a time.

    :param file input_file: FASTA file object
    :param str alphabet: Allowed characters in FASTA sequence
    :return: FASTA ID, Comment, and Sequence
    :rtype: generator
    """
    regexp_string = ''.join(['[^', alphabet, alphabet.lower(), ']+'])
    nucs_regexp = re.compile(regexp_string)  # Regexp to identify sequence lines
    first_seq = True  # Used to bypass first seq
    seq, seq_id, seq_comments = None, None, None
    for line in input_file:
        line = line.lstrip()
        if len(line) == 0:  # blank line
            continue
        if line[0] == '>':  # id line
            if first_seq:
                first_seq = False
                #  >seq_id comments becomes ['', 'seq_id','comments']
                temp_id = re.split(r'[>\s,]+', line, maxsplit=2)
                seq_id = temp_id[1]
                #  right strip to remove trailing newline character
                if len(temp_id) == 3:
                    seq_comments = temp_id[2].rstrip()
                seq = ''
                continue
            if nucs_regexp.findall(seq):
                seq = re.sub(nucs_regexp, '', seq)
            yield [seq_id, seq_comments, seq.upper()]
            #  >seq_id comments becomes ['', 'seq_id','comments']
            temp_id = re.split(r'[>\s,]+', line, maxsplit=2)
            seq_id = temp_id[1]
            if len(temp_id) == 3:
                seq_comments = temp_id[2].rstrip()
            seq = ''
            continue
        #  Remove whitespaces in sequence string
        line = ''.join(line.split())
        seq += line
    # If seq hasn't been initialized, the input file was empty.
    try:
        seq
    except UnboundLocalError:
        raise RuntimeError('Input peptides file was empty.')
    if seq:
        if nucs_regexp.findall(seq):
            seq = re.sub(nucs_regexp, '', seq)
        yield [seq_id, seq_comments, seq.upper()]


def chrom_sort(in_chroms):
    """
    Sort a list of chromosomes in the order 1..22, X, Y, M.

    :param list in_chroms: Input chromsomes
    :return: Sorted chromosomes
    :rtype: list[str]
    """
    if not in_chroms:
        return in_chroms
    chr_prefix = False
    if in_chroms[0].startswith('chr'):
        in_chroms = [x.lstrip('chr') for x in in_chroms]
        chr_prefix = True
    assert in_chroms[0] in [str(x) for x in range(1, 23)] + ['X', 'Y', 'M']
    in_chroms = sorted(in_chroms, key=lambda c: int(c) if c not in ('X', 'Y', 'M') else c)
    try:
        m_index = in_chroms.index('M')
    except ValueError:
        pass
    else:
        in_chroms.pop(m_index)
        in_chroms.append('M')
    # At this point it should be nicely sorted
    if chr_prefix:
        in_chroms = [''.join(['chr', x]) for x in in_chroms]
    return in_chroms


class GTFRecord(object):
    def __init__(self, line):
        """
        Converts a GTF record into a python object

        :param line str|list line: GTF record
        """
        if isinstance(line, str):
            line = line.strip().split('\t')
        if len(line) != 9:
            msg = '\n'.join(self.attributes)
            raise ValueError('Malformed GTF line.'
                             ' Must have the following attributes: {}'.format(msg))
        for (attr, func), value in zip(self.attributes, line):
            setattr(self, attr, func(value))
        for feature in self.attribute.split(';'):
            try:
                attr, value = feature.split()
                if attr in self.__dict__:
                    if isinstance(getattr(self, attr), list):
                        setattr(self, attr, getattr(self, attr) + [value.replace('"', '')])
                    else:
                        setattr(self, attr, [getattr(self, attr), value.replace('"', '')])
                else:
                    setattr(self, attr, value.replace('"', ''))
            except ValueError:
                pass

    attributes = [('seqname', str), ('source', str), ('feature', str), ('start', int),
                  ('end', int), ('score', str), ('strand', str), ('frame', str), ('attribute', str)]

    def __repr__(self):
        if self.feature == 'gene':
            return 'GTF({}, gene:{}, start:{}, length:{})'.format(self.seqname,
                                                                  self.gene_name,
                                                                  self.start,
                                                                  self.end - self.start + 1)
        elif self.feature == 'start_codon':
            return 'GTF({}, start codon:{}, start:{}, length:{})'.format(self.seqname,
                                                                         self.gene_name,
                                                                         self.start,
                                                                         self.end - self.start + 1)
        elif self.feature == 'transcript':
            return 'GTF({}, transcript:{}, start:{}, length:{})'.format(self.seqname,
                                                                        self.transcript_name,
                                                                        self.start,
                                                                        self.end - self.start + 1)
        elif self.feature == 'exon':
            return 'GTF({}, exon:{}, start:{}, length:{}, id:{})'.format(self.seqname,
                                                                         self.exon_number,
                                                                         self.start,
                                                                         self.end - self.start + 1,
                                                                         self.exon_id)
        elif self.feature == 'CDS':
            return 'GTF({}, CDS:{}, start:{}, length:{}, id:{})'.format(self.seqname,
                                                                         self.exon_number,
                                                                         self.start,
                                                                         self.end - self.start + 1,
                                                                         self.exon_id)

    def __hash__(self):
        if self.feature == 'gene':
            return hash(self.gene_id)
        elif self.feature == 'transcript':
            return hash(self.transcript_id)
        elif self.feature == 'CDS':
            return hash('CDS' + self.exon_id)
        elif self.feature == 'exon':
            return hash(self.exon_id)

    def __eq__(self, other):
        try:
            if self.feature == 'gene':
                return self.gene_id == other.gene_id
            elif self.feature == 'transcript':
                return self.transcript_id == other.transcript_id
            elif self.feature == 'CDS':
                if other.feature == 'CDS':
                    return self.exon_id == other.exon_id
                else:
                    return False
            elif self.feature == 'exon':
                return self.exon_id == other.exon_id
            else:
                return False
        except AttributeError:
            return False

# A named tuple describing the result of running reject_mutation on a mutation.
reject_decision = collections.namedtuple('reject_decision', (
    # The decision on whether to reject the mutation or not
    'reject',
    # The reason for rejection, if decision == True
    'reason',
    # The number of reads at the position
    'coverage',
    # The variant allele frequency
    'vaf'))

# Standard Genetic Code from NCBI
amino = 'FFLLSSSSYY**CC*WLLLLPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG'
base1 = 'TTTTTTTTTTTTTTTTCCCCCCCCCCCCCCCCAAAAAAAAAAAAAAAAGGGGGGGGGGGGGGGG'
base2 = 'TTTTCCCCAAAAGGGGTTTTCCCCAAAAGGGGTTTTCCCCAAAAGGGGTTTTCCCCAAAAGGGG'
base3 = 'TCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAGTCAG'
genetic_code = {''.join([b1, b2, b3]): aa
                for aa, b1, b2, b3 in zip(amino, base1, base2, base3)}

EFF = collections.namedtuple('EFF',
                             'annotation, '  # The annotation of the event
                             'codon_change, '  # A tuple of the changed nucleotide in the codon
                             'aa_change, '
                             'gene_name, '
                             'transcript_id, '
                             'transcript_len, '
                             'warnings'
                             )

BEDPE = collections.namedtuple('BEDPE',
                               'chrom1, start1, end1, '
                               'chrom2, start2, end2, '
                               'name, '
                               'score, '
                               'strand1, strand2, '
                               'junctionSeq1, junctionSeq2, '
                               'hugo1, hugo2')

# A dictionary to translate used legacy snpeff annotations to sequence ontology
snpeff_to_so = {'NON_SYNONYMOUS_CODING': 'missense_variant',
                'STOP_GAINED': 'stop_gained',
                'SYNONYMOUS_CODING': 'synonymous_variant',
                'FRAME_SHIFT': 'frameshift_variant',
                'CODON_CHANGE_PLUS_CODON_DELETION': 'disruptive_inframe_deletion',
                'CODON_CHANGE_PLUS_CODON_INSERTION': 'disruptive_inframe_insertion',
                'CODON_DELETION': 'inframe_deletion',
                'CODON_INSERTION': 'inframe_insertion',
                }


amino_letter = {'Ala': 'A',
                'Arg': 'R',
                'Asn': 'N',
                'Asp': 'D',
                'Cys': 'C',
                'Glu': 'E',
                'Gln': 'Q',
                'Gly': 'G',
                'His': 'H',
                'Ile': 'I',
                'Leu': 'L',
                'Lys': 'K',
                'Met': 'M',
                'Phe': 'F',
                'Pro': 'P',
                'Ser': 'S',
                'Thr': 'T',
                'Trp': 'W',
                'Tyr': 'Y',
                'Val': 'V',
                'Ter': '*',
                '*': '*'
                }


def get_exons(genome_file, annotation_file, genes_of_interest, drop_transcript_version=False):
    """
    Generates list of GTFRecord objects for each transcript and the position of the start codon
    in a separate dict

    :param file genome_file: Reference genome FASTA file
    :param file annotation_file: Genome annotation file (GTF)
    :param set(str) genes_of_interest: The genes in this sample that might need translation.
    :param bool drop_transcript_version: Drop the version part of transcript ids.
    :return: list GTFRecord of exons for each transcript, metadata for each transcript
    :rtype: dict(str, list(GTFRecord)), dict(str, int|bool)
    """
    annotation_file.seek(0)
    chroms = {}
    exons = collections.defaultdict(list)
    cds_starts = {}
    for header, comment, seq in read_fasta(genome_file, 'ACGTN'):
        chroms[header] = seq

    for line in annotation_file:
        if line.startswith('#'):
            continue
        else:
            gtf = GTFRecord(line)
            if gtf.feature == 'exon' and gtf.gene_name in genes_of_interest:
                transcript_id = gtf.transcript_id
                if drop_transcript_version:
                    transcript_id = transcript_id.split('.')[0]
                gtf.sequence = chroms[gtf.seqname][gtf.start - 1: gtf.end]
                exons[transcript_id].append(gtf)
            elif gtf.feature == 'start_codon' and gtf.gene_name in genes_of_interest:
                transcript_id = gtf.transcript_id
                if drop_transcript_version:
                    transcript_id = transcript_id.split('.')[0]
                cds_starts[transcript_id] = gtf.start
    return exons, cds_starts


def read_genes_from_gtf(gtf_file):
    """
    Read the gene annotations into a dict

    :param file gtf_file: A file handle ot the annotation file.
    :returns:  A dict of a gtf record for each gene
    :rtype: dict(string, GTFRecord)
    """
    gene_annotations = {}
    for line in gtf_file:
        if line.startswith('#'):
            continue
        else:
            gtf = GTFRecord(line)
            if gtf.feature == 'gene':
                gene_annotations[gtf.gene_name] = gtf
    return gene_annotations


def translate(seq):
    """
    Translates DNA sequence into protein sequence using globally defined genetic code

    :param str seq: DNA sequence
    :returns: Translated sequence
    :rtype: str

    >>> translate('ATGTTTCGTT')
    'MFR'
    """
    start = 0
    n = len(seq)
    codons = (seq[i: i+3] for i in range(start, n - n % 3, 3))
    protein = [genetic_code[codon] for codon in codons]
    return ''.join(protein)


def first_mismatch(str1, str2):
    """
    Returns the position of the first mismatch between 2 strings.

    :param str1: The first string
    :param str2: The second string
    :return: Position of the first mismatch else -1 if no mismatches

    >>> first_mismatch('abc', 'def')
    0
    >>> first_mismatch('abc', 'abe')
    2
    >>> first_mismatch('abc', 'abc')
    -1
    >>> first_mismatch('abc', 'abcde')
    -1
    """
    for i, j in enumerate(zip(str1, str2)):
        if j[0] != j[1]:
            return i
    else:
        return -1


def is_gzipfile(filename):
    """
    Attempt to ascertain the gzip status of a file based on the "magic signatures" of the file.

    This was taken from the stack overflow post
    http://stackoverflow.com/questions/13044562/python-mechanism-to-identify-compressed-file-type\
        -and-uncompress

    :param str filename: A path to a file
    :return: True if the file appears to be gzipped else false
    :rtype: bool
    """
    assert os.path.exists(filename), 'Input {} does not point to a file.'.format(filename)
    with open(filename, 'rb') as in_f:
        start_of_file = in_f.read(3)
        if start_of_file == '\x1f\x8b\x08':
            return True
        else:
            return False


def file_type(filename):
    """
    This module is used to open an input file in the appropriate method dependiing on whether it is a
    regular file, a gzipped
    file or a url.

    :param str filename: The path to the file
    :return: an open file handle to the file
    :rtype: file
    """
    if is_gzipfile(filename):
        return gzip.GzipFile(filename, 'r')
    else:
        return open(filename, 'r')


def get_snpeff_3_6_changes(eff):
    """
    Get the changes for the call from a snpEff 3.6 annotated vcf.

    :param str eff: The EFF field in the INFO column
    :return: The Effects of the mutation on the given protein
    :rtype: list(EFF)
    """
    out_effs = []
    eff = re.sub('EFF=', '', eff)
    for e in eff.split(','):
        if 'protein_coding' in e:
            if e.startswith(('NON_SYNONYMOUS_CODING', 'SYNONYMOUS_CODING', 'STOP_GAINED')):
                e = e.strip().split('|')
                out_effs.append(EFF(annotation=snpeff_to_so[e[0].split('(')[0]],
                                    codon_change=e[2],
                                    aa_change=e[3],
                                    gene_name=e[5],
                                    transcript_id=e[8],
                                    transcript_len=e[4],
                                    warnings=e[-1].rstrip(')') if len(e) == 12 else ''))
            elif e.startswith(('CODON', 'FRAME')):
                e = e.strip().split('|')
                out_effs.append(EFF(annotation=snpeff_to_so[e[0].split('(')[0]],
                                    codon_change='',
                                    aa_change=e[3],
                                    gene_name=e[5],
                                    transcript_id=e[8],
                                    transcript_len=e[4],
                                    warnings=e[-1].rstrip(')') if len(e) == 12 else ''))
    return out_effs


def get_snpeff_4_0_changes(eff):
    """
    Get the changes for the call from a snpEff 4.0 annotated vcf.

    :param str eff: The EFF field in the INFO column
    :return: The Effects of the mutation on the given protein
    :rtype: list(EFF)
    """
    out_effs = []
    eff = re.sub('EFF=', '', eff)
    for e in eff.split(','):
        if 'protein_coding' in e:
            if e.startswith(('missense_variant', 'synonymous_variant', 'stop_gained')):
                e = e.strip().split('|')
                out_effs.append(EFF(annotation=e[0].split('(')[0],
                                    codon_change=e[2],
                                    aa_change=translate_aa_change_4_0(e[3]),
                                    gene_name=e[5],
                                    transcript_id=e[8],
                                    transcript_len=e[4],
                                    warnings=e[-1].rstrip(')') if len(e) == 12 else ''))
            elif e.startswith(('frame', 'inframe', 'disruptive')):
                e = e.strip().split('|')
                # Sinve SNPEff 4.0 does things differently, we will handle all indels similarly to
                # frameshifts (otherwise we will predict the wrong things based on the annotation.
                # The only downside to this is that we will predict excessively long peptides if the
                # user specified --indel_extend_length > n
                aa_change = list(translate_aa_change_4_0(e[3]))
                while True:
                    temp = aa_change.pop()
                    if temp.isdigit():
                        aa_change.append(temp + '?')
                        aa_change = ''.join(aa_change)
                        break
                annotation = 'frameshift_variant'
                out_effs.append(EFF(annotation=annotation,
                                    codon_change='',
                                    aa_change=aa_change,
                                    gene_name=e[5],
                                    transcript_id=e[8],
                                    transcript_len=e[4],
                                    warnings=e[-1].rstrip(')') if len(e) == 12 else ''))
    return out_effs


def translate_aa_change_4_0(aa_change_field):
    """
    Translate an amino acid change in snpeff 4.0 to 3.6 form.

    :param str aa_change_field: The field containing the aa change
    :return: The changed amino acid(s) in the for <REF><POS><ALT>
    :rtype: str

    >>> translate_aa_change_4_0('p.Gly144_Cys145del/c.431_433delGCT')
    'GC144'
    >>> translate_aa_change_4_0('p.Asp479_Ser480fs')
    'DS479?'
    >>> translate_aa_change_4_0('p.His87_Ala88insGlnLeu')
    'HA87QL'
    >>> translate_aa_change_4_0('p.Gly100Arg')
    'G100R'
    >>> translate_aa_change_4_0('p.Trp102fs')
    'W102?'
    >>> translate_aa_change_4_0('p.Tyr17*')
    'Y17*'
    """
    if '/' in aa_change_field:
        aa_change_field = aa_change_field.split('/')[0]

    aa_change_field = aa_change_field[2:]  # remove the leading "p."

    if any(x in aa_change_field for x in ['ins', 'del', 'fs']):
        # Indel
        aa_change_field = aa_change_field.split('_')

        # handle the last one first, then move towards the first
        prefix, aa_change_field[-1] = amino_letter[aa_change_field[-1][:3]], aa_change_field[-1][3:]
        pos = aa_change_field[-1].translate(None, string.letters)
        aa_change_field[-1] = aa_change_field[-1].translate(None, string.digits)

        if aa_change_field[-1] == 'fs':
            suffix = '?'
        else:
            suffix = ''.join([amino_letter[aa_change_field[-1][x:x + 3]]
                              for x in range(0, len(aa_change_field[-1]), 3)
                              if aa_change_field[-1][x:x + 3] not in ['del', 'ins']])

        aa_change_field.pop()
        for change in aa_change_field:
            prefix = amino_letter[change[:3]] + prefix
            pos = change[3:]
    else:
        prefix, aa_change_field = amino_letter[aa_change_field[:3]], aa_change_field[3:]
        if aa_change_field.endswith('*'):
            pos, suffix = aa_change_field[:-1], '*'
        else:
            pos, suffix = aa_change_field[:-3], amino_letter[aa_change_field[-3:]]

    return prefix + pos + suffix


def get_snpeff_4_x_changes(eff):
    """
    Get the changes for the call from a snpEff 4.1+ annotated vcf.

    :param str eff: The ANN field in the INFO column
    :return: The Effects of the mutation on the given protein
    :rtype: list(EFF)
    """
    out_effs = []
    eff = re.sub('ANN=', '', eff)
    for e in eff.split(','):
        if 'protein_coding' in e:
            e = e.strip().split('|')
            if e[1] in ('missense_variant', 'synonymous_variant', 'stop_gained'):
                out_effs.append(EFF(annotation=e[1],
                                    codon_change=e[9],
                                    aa_change=translate_aa_change_4_x(e[10]),
                                    gene_name=e[3],
                                    transcript_id=e[6],
                                    transcript_len=e[13].split('/')[1],
                                    warnings=e[-1].rstrip(')') if len(e) == 16 else ''))
            elif any(x in e[1] for x in ('frame', 'inframe', 'disruptive')):
                out_effs.append(EFF(annotation=e[1],
                                    codon_change=e[9],
                                    aa_change=translate_aa_change_4_x(e[10]),
                                    gene_name=e[3],
                                    transcript_id=e[6],
                                    transcript_len=e[13].split('/')[1],
                                    warnings=e[-1].rstrip(')') if len(e) == 16 else ''))
    return out_effs


def translate_aa_change_4_x(aa_change_field):
    """
    Translate an amino acid change in snpeff 4.1+ to 3.6 form.

    :param str aa_change_field: The field containing the aa change
    :return: The changed amino acid(s) in the for <REF><POS><ALT>
    :rtype: str

    >>> translate_aa_change_4_x('p.Gly144_Cys145del/c.431_433delGCT')
    'GC144'
    >>> translate_aa_change_4_x('p.Asp479_Ser480fs')
    'DS479?'
    >>> translate_aa_change_4_x('p.His87_Ala88insGlnLeu')
    'A88QLA'
    >>> translate_aa_change_4_x('p.Gly100Arg')
    'G100R'
    >>> translate_aa_change_4_x('p.Trp102fs')
    'W102?'
    >>> translate_aa_change_4_x('p.Tyr17*')
    'Y17*'
    >>> translate_aa_change_4_x('p.Tyr174_Arg175delinsTrp')
    'YR174W'
    >>> translate_aa_change_4_x('p.Arg175delinsTrpLeu')
    'R175WL'
    """
    if '/' in aa_change_field:
        aa_change_field = aa_change_field.split('/')[0]

    aa_change_field = aa_change_field[2:]  # remove the leading "p."

    if any(x in aa_change_field for x in ['ins', 'del', 'fs']):
        # Indel
        aa_change_field = aa_change_field.split('_')

        # handle the last one first, then move towards the first
        prefix, aa_change_field[-1] = amino_letter[aa_change_field[-1][:3]], aa_change_field[-1][3:]
        pos = aa_change_field[-1].translate(None, string.letters)
        aa_change_field[-1] = aa_change_field[-1].translate(None, string.digits)

        if aa_change_field[-1] == 'fs':
            suffix = '?'
        else:
            suffix = ''
            if aa_change_field[-1].startswith('delins'):
                aa_change_field[-1] = aa_change_field[-1][6:]  # Strip "delins"
            elif aa_change_field[-1] == 'del':
                aa_change_field[-1] = ''  # Strip "del"
            else:
                assert aa_change_field[-1].startswith('ins')
                aa_change_field = [aa_change_field[-1][3:]]  # Strip "ins"
                suffix = prefix  # Make it so the results will be C>ABC
            suffix = ''.join([amino_letter[aa_change_field[-1][x:x + 3]]
                              for x in range(0, len(aa_change_field[-1]), 3)]) + suffix
        aa_change_field.pop()
        for change in aa_change_field:
            prefix = amino_letter[change[:3]] + prefix
            pos = change[3:]
    else:
        prefix, aa_change_field = amino_letter[aa_change_field[:3]], aa_change_field[3:]
        if aa_change_field.endswith('*'):
            pos, suffix = aa_change_field[:-1], '*'
        else:
            pos, suffix = aa_change_field[:-3], amino_letter[aa_change_field[-3:]]

    return prefix + pos + suffix


def get_vep_changes(eff):
    """
    Get the changes for the call from a VEP annotated vcf.

    :param str eff: The CSQ field in the INFO column
    :return: The Effects of the mutation on the given protein
    :rtype: list(EFF)
    """
    out_effs = []
    eff = re.sub('ANN=', '', eff)
    for e in eff.split(','):
        if 'protein_coding' in e:
            e = e.strip().split('|')
            if e[1] in ('missense_variant', 'synonymous_variant', 'stop_gained'):
                out_effs.append(EFF(annotation=e[1],
                                    codon_change=e[16],
                                    aa_change=translate_aa_change_vep(e[14], e[15]),
                                    gene_name=e[3],
                                    transcript_id=e[6],
                                    transcript_len=-1,
                                    warnings=''))
            elif any(x in e[1] for x in ('frame', 'inframe', 'protein_altering')):
                out_effs.append(EFF(annotation=e[1],
                                    codon_change=e[16],
                                    aa_change=translate_aa_change_vep(e[14], e[15]),
                                    gene_name=e[3],
                                    transcript_id=e[6],
                                    transcript_len=-1,
                                    warnings=''))
    return out_effs


def translate_aa_change_vep(pos_field, aa_change_field):
    """
    Translate an amino acid change in snpeff 4.1+ to 3.6 form.

    :param str aa_change_field: The field containing the aa change
    :return: The changed amino acid(s) in the for <REF><POS><ALT>
    :rtype: str

    >>> translate_aa_change_vep('144-145', 'GC/-')
    'GC144'
    >>> translate_aa_change_vep('479-480', 'DS/X')
    'DS479?'
    >>> translate_aa_change_vep('87-88', '-/QL')
    '-87QL'
    >>> translate_aa_change_vep('100', 'G/R')
    'G100R'
    >>> translate_aa_change_vep('102-103', 'W/X')
    'W102?'
    >>> translate_aa_change_vep('17', 'Y/*')
    'Y17*'
    >>> translate_aa_change_vep('174', 'YR/W')
    'YR174W'
    >>> translate_aa_change_vep('175', 'R/WL')
    'R175WL'
    >>> translate_aa_change_vep('175', 'R')
    'R175R'
    """
    aa_change_field = aa_change_field.split('/')
    if len(aa_change_field) == 1:
        # Synonymous mutation
        aa_change_field = aa_change_field + aa_change_field

    pos_field = pos_field.split('/')[0].split('-')[0]
    if aa_change_field[1].endswith('X'):
        aa_change_field[1] = '?'
    elif aa_change_field[1] == '-':
        aa_change_field[1] = ''
    prefix, pos, suffix = aa_change_field[0], pos_field, aa_change_field[1]
    return prefix + pos + suffix
